/*

  DBOX Module
  Copyright (C) 2003 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <iostream>

#include <stringtools.h>
#include <libdbox/md5.h>
#include <dblib/str_getline.h>
#include <dblib/index.h>
#include <dblib/getpeer.h>
#include <dbclient/auth.h>
#include <dbclient/logstr.h>

class dbauthdt
 {
 public:
  dbauthdt(std::istream &_in, std::ostream &_out);
  
  void dbauthd();

 protected:
  std::istream &in;
  std::ostream &out;

  std::string peer;

  indext auth_index;

  void authenticate(const std::string &address,
                    const std::string &user,
                    const std::string &error);

  void status_response(unsigned code, const std::string &msg);

  void parse_command();

  bool read(std::string &s);
 };

/*******************************************************************\

Function: dbauthdt::authenticate

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbauthdt::authenticate(const std::string &service,
                            const std::string &username,
                            const std::string &password)
 {
  autht auth;

  try
   {
    auth.auth(username, password);
    status_response(100, "autentication successfull");
    logstr << service << ": authenticated " << username;
    logstr.log("authenticate");
   }

  catch(auth_excpt e)
   {
    logstr << service << ": authentication of user "
           << username << " failed";
    logstr.log("authenticate");
    status_response(400, e.what());
   }
 }

/*******************************************************************\

Function: dbauthdt::dbauthdt

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

dbauthdt::dbauthdt(std::istream &_in, std::ostream &_out):
  in(_in), out(_out)
 {
 }

/*******************************************************************\

Function: dbauthdt::dbauthdt

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbauthdt::dbauthd()
 {
  try
   {  
    peer=getpeer(0);

    std::string service, username, password;

    if(read(service)) return;
    if(read(username)) return;
    if(read(password)) return;

    authenticate(service, username, password);
   }

  catch(...)
   {
    status_response(450, "internal server error");
   }
 }

/*******************************************************************\

Function: dbauthdt::read_command 

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbauthdt::read(std::string &s)
 {
  bool got_timeout;

  if(!str_getline(in, s, true, 60, got_timeout))
    return true;

  if(got_timeout)
   {
    status_response(451, "timeout waiting for input");
    logstr << "timeout while talking to " << peer;
    logstr.log("read");
    return true;
   }

  return false;
 }

/*******************************************************************\

Function: dbauthdt::status_response

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbauthdt::status_response(unsigned code, const std::string &msg)
 {
  out << code << " " << msg << "\r\n";
 }

/*******************************************************************\

Function: dbauthd

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbauthd()
 {
  dbauthdt dbauthd(std::cin, std::cout);

  dbauthd.dbauthd();
 }
