#include <signal.h>

#include <algorithm>
#include <vector>

#include <libdbox/stringtools.h>
#include <libdbox/general.h>

#include "checkfilename.h"

#include <dblib/os_dep_file.h>
#include <dblib/os_dep_socket.h>
#include <dbclient/logstr.h>
#include <dbclient/client.h>
#include <dbclient/user.h>
#include <dbclient/auth.h>

#include "dbftpd.h"

/*******************************************************************\

Function: dbftpdt::docommand

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbftpdt::docommand(const std::string &command)
 {
  std::string args, cmd;

   {
    std::string::size_type pos=command.find(' ');

    if(pos!=std::string::npos)
     {
      cmd.assign(command, 0, pos);

      while(command[pos]==' ' && pos<command.size()) pos++;
      args.assign(command, pos, command.size()-pos);
     }
    else
      cmd=command;
   }      

  if(cmd=="USER")
   {
    if(douser(args)) return TRUE;
   }
  else if(cmd=="PASS")
   {
    if(dopass(args)) return TRUE;
   }
  else if(cmd=="QUIT")
   {
    out << "221 Goodbye.\r\n";

    if(login)
      logstr << username << "/" << alias << ": Logout";
    else
      logstr << "Logout";

    logstr.log("dboxftpd");

    return TRUE;
   }
  else if(cmd=="PORT")
   {
    if(login)
     {
      if(doport(args))
       {
        out << "425 Internal error\r\n";
        return TRUE;
       }
     }
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="PASV")
   {
    if(login)
     {
      if(dopasv())
       {
        out << "425 Internal error\r\n";
        return TRUE;
       }
     }
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="TYPE")
   {
    out << "200 Type is ignored\r\n";
   }
  else if(cmd=="STRU")
   {
    out << "200 Structure file is assumed\r\n";
   }
  else if(cmd=="MODE")
   {
    if(toupper(args[0])=='S')
      out << "200 MODE S ok.\r\n";
    else
     {
      out << "502 Unimplemented MODE type.\r\n";
      logstr << "MODE " << args << " not implemented";
      logstr.log();
     }
   }
  else if(cmd=="RETR")
   {
    if(login)
     {
      logstr << "RETR not implemented";
      logstr.log("dboxftpd");
      out << "502 not implemented\r\n";
     }
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="STOR")
   {
    if(login)
      dostor(args);
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="APPE")
   {
    if(login)
     {
      logstr << "APPE not implemented";
      logstr.log("dboxftpd");
      out << "502 not implemented\r\n";
     }
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="ALLO")
   {
    out << "200 nooping...\r\n";
   }
  else if(cmd=="REST")
   {
    if(login)
     {
      logstr << "REST not implemented";
      logstr.log("dboxftpd");
      out << "502 not implemented\r\n";
     }
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="RNFR")
   {
    if(login)
     {
      logstr << "RNFR not implemented";
      logstr.log("dboxftpd");
      out << "502 not implemented\r\n";
     }
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="RNTO")
   {
    if(login)
     {
      logstr << "RNTO not implemented";
      logstr.log("dboxftpd");
      out << "502 not implemented\r\n";
     }
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="ABOR")
   {
    if(login)
     {
      logstr << "ABOR not implemented";
      logstr.log("dboxftpd");
      out << "502 not implemented\r\n";
     }
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="DELE")
   {
    if(login)
      dodele(args);
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="RMD")
   {
    if(login)
      dodele(args);
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="MKD")
   {
    if(login)
      domkd(args);
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="PWD")
   {
    if(login)
      out << "257 \"/" << cwd << "\" is current directory\r\n";
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="LIST")
   {
    if(login)
      dolist(FALSE, args);
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="NLST")
   {
    if(login)
      dolist(TRUE, args);
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="SITE")
   {
    if(login)
     {
      logstr << "SITE not implemented";
      logstr.log("dboxftpd");
      out << "502 not implemented\r\n";
     }
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="SYST")
   {
    out << "215 UNIX Type: L8\r\n";
   }
  else if(cmd=="STAT")
   {
    logstr << "STAT not implemented";
    logstr.log("dboxftpd");
    out << "502 not implemented\r\n";
   }
  else if(cmd=="HELP")
   {
    out << "214 Sorry, no help\r\n";
   }
  else if(cmd=="CDUP")
   {
    if(login)
      docwd("..");
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="CWD")
   {
    if(login)
      docwd(args);
    else
      out << "530 Please login first.\r\n";
   }
  else if(cmd=="NOOP")
   {
    out << "200 nooping...\r\n";
   }
  else
   {
    out << "502 not implemented\r\n";
   }

  return FALSE;
 }

/*******************************************************************\

Function: ftpalarm

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void ftpalarm(int signal)
 {
  std::cout << "221 Timeout\r\n";
  exit(1);
 }

/*******************************************************************\

Function: dboxftpd::doit

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpdt::doit()
 {
  try
   {
    init();

    if(homepagedir=="")
     {
      out << "502 Service not configured\r\n";
      return;
     }

    welcome();
    main_loop();
   }

  catch(client_excpt &e)
   {
    out << "502 Server failure - service not available.\r\n";
   }
 }

/*******************************************************************\

Function: dboxftpd::init

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpdt::init()
 {
  homepagedir=client.config["homepage.dir"];

  failcount=3;
  login=has_homedir=FALSE;
  passive_socket=FALSE;
  port_sa_ok=FALSE;
   
   {
    socklen_t len;

    len=sizeof(peer);
    if(getpeername(0, (struct sockaddr *)&peer, &len)<0)
      return;
   }

  signal(SIGALRM, ftpalarm);
 }

/*******************************************************************\

Function: dboxftpd::welcome

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpdt::welcome()
 {
  out << "220-All transfers are logged.\r\n"
         "220-\r\n"
         "220 "
      << client.config["hostname"]
      << "."
      << client.config["domain"]
      << " DBOX FTP Server ready\r\n";

  out.flush();
  fflush(stdout);
 }

/*******************************************************************\

Function: dboxftpd::main_loop

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpdt::main_loop()
 {
  while(TRUE)
   {
    char command[TEXTLEN];

    alarm(10*60);
    if(fgets(command, TEXTLEN-1, stdin)==NULL) break;
    alarm(0);

    char *tptr;
    tptr=strchr(command, '\r'); if(tptr!=NULL) *tptr=0;
    tptr=strchr(command, '\n'); if(tptr!=NULL) *tptr=0;

    if(docommand(command)) return;

    fflush(stdout);
   }
 }

/*******************************************************************\

Function: dbftpd

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpd()
 {
  dbftpdt dbftp(std::cin, std::cout);

  dbftp.doit();
 }
