#include <algorithm>
#include <iostream>
#include <vector>

#include <libdbox/stringtools.h>
#include <dblib/os_dep_file.h>
#include <dblib/os_dep_socket.h>
#include <dbclient/logstr.h>
#include <dbclient/client.h>

#include "checkfilename.h"
#include "dbftpd.h"
#include "www_table.h"

/*******************************************************************\

Function: dbftpdt::get_filename

  Inputs: full path

 Outputs: filename

 Purpose: extract filename out of full path

\*******************************************************************/

std::string dbftpdt::get_filename(const std::string &full_path)
 {
  const char *file=strrchr(full_path.c_str(), DIRCHAR[0]);

  if(file==NULL)
    return full_path;

  return file++;
 }

/*******************************************************************\

Function: dbftpdt::mk_filename

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpdt::mk_filename(const std::string &args, std::string &filename)
 {
  if(args.size()!=0 && args[0]=='/')
    filename=homedir+(args.c_str()+1);
  else if(cwd=="")
    filename=homedir+args;
  else
    filename=homedir+cwd+DIRCHAR+args;
 }

/*******************************************************************\

Function: dbftpdt::mk_homedir

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpdt::mk_homedir()
 {
  if(has_homedir) return;

  std::string filename=homepagedir+alias;

  if(mkdir(filename.c_str(), 0755)!=0)
   {
    logstr << "failed to create " << filename;
    logstr.log("mk_homedir");
   }
  else
   {
    logstr << "created " << filename;
    logstr.log("mk_homedir");
   }

  has_homedir=TRUE;
 }

/*******************************************************************\

Function: dbftpdt::dcon_open

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbftpdt::dcon_open()
 {
  socklen_t len;
  int s, on, trials;

  if(passive_socket)
   {
    len=sizeof(sa);
    s=accept(data_fd, (struct sockaddr *)&sa, &len);
    if(s<0 || ntohs(sa.sin_port)<IPPORT_RESERVED ||
       sa.sin_addr.s_addr!=peer.sin_addr.s_addr)
     {
      out << "425 accept failed.\r\n";
      return TRUE;
     }

    close(data_fd);
    data_fd=s;
   }
  else
   {
    if(!port_sa_ok)
     {
      logstr << username << ": tried to make default active connection";
      logstr.log("dcon_open");
      out << "500 not implemented.\r\n";
      return TRUE;
     }

    data_fd=socket(AF_INET, SOCK_STREAM, 0);
    if(data_fd<0) return TRUE;

    setsockopt(data_fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

    trials=3;
   loop:
    if(connect(data_fd, (struct sockaddr *)&port_sa, sizeof(sa))<0)
     {
      if((errno==EAGAIN || errno==EINPROGRESS) && trials>0)
       {
        trials--;
        sleep(1);
        goto loop;
       } 
      return TRUE;
     }

    fflush(stdout);
   }

  return FALSE;
 }

/*******************************************************************\

Function: dbftpdt::dcon_close

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpdt::dcon_close()
 {
  close(data_fd);
  passive_socket=FALSE;
 }

/*******************************************************************\

Function: dbftpdt::domkd

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpdt::domkd(const std::string &args)
 {
  std::string filename;

  if(strstr(args.c_str(), "..")!=NULL)
   {
    out << "504 Not available for security reasons.\r\n";
    return;
   }

  if(!has_homedir)
    mk_homedir();

  mk_filename(args, filename);

  if(checkfilename(get_filename(filename).c_str()))
   {
    out << "550 Illegal file name\r\n";
    logstr << username << ": tried to mkdir " << filename;
    logstr.log("dboxftpd::domkd");
    return;
   }

  if(mkdir(filename.c_str(), 0755)!=0)
   {
    out << "550 Failed to create\r\n";
    logstr << username << ": failed to mkdir " << filename;
    logstr.log("dboxftpd::domkd");
    return;
   }

  out << "257 MKD command successful.\r\n";

  logstr << username << ": mkdir " << filename;
  logstr.log("dboxftpd::domkd");
 }

/*******************************************************************\

Function: dbftpdt::dodele

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpdt::dodele(const std::string &args)
 {
  std::string filename;

  if(strstr(args.c_str(), "..")!=NULL)
   {
    out << "504 Not available for security reasons.\r\n";
    return;
   }

  if(!has_homedir) 
   {
    out << "550 Failed to delete\r\n";
    return;
   }

  mk_filename(args, filename);

  if(checkfilename(get_filename(filename).c_str()))
   {
    out << "550 Illegal file name\r\n";
    logstr << username << ": tried to delete " << filename;
    logstr.log("dboxftpd::dodele");
    return;
   }

  #if 0
  www_tablet www_table(alias);
  if(!www_table.lock("dbftpdt::dodele"))
   {
    www_quota(www_table, filename.c_str(), FALSE);
    www_table.write();
    www_table.unlock();
   }
  #endif

  if(unlink(filename.c_str())!=0)
   {
    if(rmdir(filename.c_str())!=0)
     {
      out << "550 Failed to delete\r\n";
      logstr << username << ": failed to delete " << filename;
      logstr.log("dboxftpd::dodele");
      return;
     }
   }

  out << "250 DELE command successful.\r\n";

  logstr << username << ": rm " << filename;
  logstr.log("dboxftpd::dodele");
 }

/*******************************************************************\

Function: dbftpdt::doport

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbftpdt::doport(const std::string &args)
 {
  char *a, *p;
  int h1, h2, h3, h4, p1, p2;
  socklen_t len;

  if(passive_socket)
   {
    close(data_fd);
    passive_socket=FALSE;
   }

  len=sizeof(port_sa);
  if(getsockname(0, (struct sockaddr *)&port_sa, &len)<0)
    return TRUE;

  a=(char *)&port_sa.sin_addr.s_addr;
  p=(char *)&port_sa.sin_port;

  sscanf(args.c_str(), "%d,%d,%d,%d,%d,%d", &h1, &h2, &h3, &h4, &p1, &p2);

  a[0]=h1&0xff;
  a[1]=h2;
  a[2]=h3;
  a[3]=h4;
  p[0]=p1;
  p[1]=p2;

  if(port_sa.sin_addr.s_addr!=peer.sin_addr.s_addr ||
     ntohs(port_sa.sin_port)<IPPORT_RESERVED)
   {
    out << "500 Illegal PORT Command "
        << args << " " 
        << h1 << "," << h2 << "," << h3 << ","
        << h4 << "," << p1 << "," << p2 << "\r\n";
    port_sa_ok=FALSE;
   }
  else
   {
    out << "200 PORT command successful.\r\n";
    port_sa_ok=TRUE;
   }

  return FALSE;
 }

/*******************************************************************\

Function: dbftpdt::docwd2

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbftpdt::docwd2(const std::string &args)
 {
  if(args=="" || args==".")
   {
   }
  else if(args=="/")
    cwd="";
  else if(args=="..")
   {
    if(cwd=="") return TRUE;

    std::string::size_type pos=cwd.rfind(cwd, DIRCHAR[0]);

    if(pos!=std::string::npos)
     {
      std::string tmp(cwd, 0, pos);
      cwd.swap(tmp);
     }
    else
      cwd="";
   }
  else if(args[0]=='/') 
   {
    cwd="";
    return docwd2(args.c_str()+1);
   }
  else
   {
    std::string::size_type pos=args.find('/');

    if(pos!=std::string::npos)
     {
      std::string first(args, 0, pos),
                  second(args, pos+1);

      if(docwd2(first)) return TRUE;
      return docwd2(second);
     }
    else
     {
      struct stat st;

      if(stat(args.c_str(), &st)<0 || !S_ISDIR(st.st_mode))
        return TRUE;

      if(cwd=="")
        cwd=args;
      else
       {
        cwd+=DIRCHAR;
        cwd+=args;
       }
     }
   }

  return FALSE;
 }

/*******************************************************************\

Function: dbftpdt::docwd

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpdt::docwd(const std::string &args)
 {
  if(docwd2(args))
    out << "550 " << args << ": No such file or directory.\r\n";
  else
    out << "250 CWD command successfull.\r\n";
 }

/*******************************************************************\

Function: dbftpdt::dostor

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpdt::dostor(const std::string &args)
 {
  struct stat st;
  std::string filename;
  int fd, count;
  char buf[4096];
  dword size=0;

  if(strstr(args.c_str(), "..")!=NULL)
   {
    out << "504 Not available for security reasons.\r\n";
    return;
   }

  if(!has_homedir)
    mk_homedir();

   {
    #if 0
    long size, files, max_size, max_files;

    get_www_quota_usage(status.userprofile, files, size);
    get_max_quota(status.userprofile, status.gruppe, max_files, max_size);

    if(max_files!=0 && files>max_files)
     {
      out << "504: File quota for " << username << " exceeded\r\n";
      return;
     }

    if(max_size!=0 && size>max_size)
     {
      out << "504: Size quota for " << username << " exceeded\r\n";
      return;
     }
    #endif
   }

  mk_filename(args, filename);

  if(checkfilename(get_filename(filename).c_str()))
   {
    out << "550 Illegal file name\r\n";
    logstr << username << ": tried to store " << filename;
    logstr.log("dboxftpd::dostor");
    return;
   }

  if(stat(filename.c_str(), &st)>=0 && !S_ISREG(st.st_mode))
   {
    out << "550 Not a regular file\r\n";
    logstr << username << ": tried to store " << filename;
    logstr.log("dboxftpd::dostor");
    return;
   }

  #if 0
  www_tablet www_table(alias);

  if(access(filename.c_str(), 00)==0)
    if(!www_table.lock("dboxftpd::dostor"))
     {
      www_quota(www_table, filename.c_str(), FALSE);
      www_table.write();
      www_table.unlock();
     }
  #endif

  fd=open(filename.c_str(), O_WRONLY|O_CREAT|O_NOCTTY|O_TRUNC|O_BINARY, 0644);

  if(fd<0)
   {
    out << "550 Failed to open\r\n";
    logstr << username << ": failed to store " << filename;
    logstr.log("dboxftpd::dostor");
    return;
   }

  out << "150 Opening BINARY mode data connection for "
      << args << ".\r\n";

  fflush(stdout);

  if(dcon_open()) return;

  while((count=read(data_fd, buf, sizeof(buf)))>0)
   {
    if(size<client.config.get_int("ftpd.maxfilesize"))
     {
      write(fd, buf, count);
      if(count>0) size+=count;
     }
   }

  close(fd);

  if(count<0)
    out << "426 Data Connection\r\n";
  else
    out << "226 Transfer complete.\r\n";

  #if 0
  if(!www_table.lock("dboxftpd::dostor"))
   {
    www_quota(www_table, filename.c_str(), TRUE);
    www_table.write();
    www_table.unlock();
   }
  #endif

  logstr << username << ": stored " << filename
         << " (" << size << "bytes)";

  logstr.log("dboxftpd::dostor");

  dcon_close();
 }

/*******************************************************************\

Function: dbftpdt::dolist_file

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpdt::dolist_file(FILE *outfile, bool nlst, const std::string &filename)
 {
  std::string file=get_filename(filename);

  if(file=="profile") return;

  if(!nlst)
   {
    struct stat st;

    if(stat(filename.c_str(), &st)>=0 && 
       (S_ISREG(st.st_mode) || S_ISDIR(st.st_mode)))
     {
      if(S_ISDIR(st.st_mode))
        fprintf(outfile, "drwxr-xr-x");
      else
        fprintf(outfile, "-rw-r--r--");

      time_t t;
      int year;
      struct tm *tm;
      char date[TEXTLEN];

      time(&t);
      tm=localtime(&t);
      year=tm->tm_year;
      tm=localtime(&st.st_mtime);

      strftime(date, TEXTLEN-1,
        (tm->tm_year==year)?"%b %d %H:%M":"%b %d  %Y", tm);

      fprintf(outfile, " %3d %-8s %-8s %8lu %12s %s\r\n",
        st.st_nlink, username.c_str(), "none", st.st_size, date,
        file.c_str());
     }
   }
  else
   {
    struct stat st;

    if(stat(filename.c_str(), &st)>=0 && S_ISREG(st.st_mode))
      fprintf(outfile, "%s\r\n", file.c_str());
   }
 }

/*******************************************************************\

Function: dbftpdt::dolist

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbftpdt::dolist(bool nlst, const std::string &args)
 {
  if(strstr(args.c_str(), "..")!=NULL)
   {
    out << "504 Not available for security reasons.\r\n";
    return;
   }

  out << "150 Opening data connection.\r\n";
  fflush(stdout);

  if(dcon_open()) return;

  if(has_homedir)
   {
    FILE *outfile=fdopen(data_fd, "wt");
    if(outfile==NULL) return;

    std::string filename;

    if(args[0]=='-')
     {
      nlst=FALSE;
      mk_filename("", filename);
     }
    else
      mk_filename(args, filename);

    struct stat st;
    std::vector<std::string> files;

    if(stat(filename.c_str(), &st)>=0)
     {
      if(S_ISREG(st.st_mode))
       {
        dolist_file(outfile, nlst, filename);
       }
      else if(S_ISDIR(st.st_mode))
       {
        DIR *dirp;

        if((dirp=opendir(filename.c_str()))!=NULL)
         {
          struct dirent *dir;

          while((dir=readdir(dirp))!=NULL)
            files.push_back(filename+DIRCHAR+dir->d_name);
	
          closedir(dirp);
         }
       }
     }

    std::sort(files.begin(), files.end());

    for(std::vector<std::string>::const_iterator it=files.begin();
        it!=files.end(); it++)
      dolist_file(outfile, nlst, *it);

    fclose(outfile);
   }

  out << "226 Transfer complete.\r\n";

  dcon_close();
 }

/*******************************************************************\

Function: dbftpdt::dopasv

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbftpdt::dopasv()
 {
  char *a, *p;
  socklen_t len;

  if(passive_socket)
   {
    close(data_fd);
    passive_socket=FALSE;
   }

  data_fd=socket(AF_INET, SOCK_STREAM, 0);
  if(data_fd<0) return TRUE;

  len=sizeof(sa);
  if(getsockname(0, (struct sockaddr *)&sa, &len)<0)
    return TRUE;

  sa.sin_port=htons(0);

  if(bind(data_fd, (struct sockaddr *)&sa, sizeof(sa))<0)
    return TRUE;

  if(getsockname(data_fd, (struct sockaddr *)&sa, &len)<0)
    return TRUE;

  if(listen(data_fd, 1)<0)
    return TRUE;

  a=(char *)&sa.sin_addr.s_addr;
  p=(char *)&sa.sin_port;

  printf("227 Entering Passive Mode (%d,%d,%d,%d,%d,%d)\r\n",
    a[0]&0xff, a[1]&0xff, a[2]&0xff, a[3]&0xff, 
    p[0]&0xff, p[1]&0xff);

  passive_socket=TRUE;

  return FALSE;
 }

