/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <libdbox/userdb.h>
#include <libdbox/md5.h>
#include <libdbox/general.h>
#include <libdbox/flags.h>
#include <dblib/base64str.h>

#include "auth.h"
#include "userlist.h"

/*******************************************************************\

Function: auth

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void do_auth(std::ostream &out, const std::string &user)
 {
  userpt up(user.c_str());

  if(up.read())
   {
    std::cerr << "failed to read profile of user "
              << user << std::endl;
    return;
   }

  const char *cryptedpassword;
  char temppw[TEXTLEN];

  cryptedpassword=up|"cryptedpassword";
  if(cryptedpassword[0]==0)
    cryptedpassword=up|"cryptedpasswort";

  if(cryptedpassword[0]==0 ||
     strcmp(cryptedpassword, "locked")==0)
    return;

  userpt up2;
  grouppt gp;

  get_profiles(up, up2, gp);

  if((up|"alias")[0]!=0)
    if(!atoi(up|"account"))
      return;

  if(!get_internalflag(up, up2, gp, LOGIN_WEB))
    return;

  if(strcmp(cryptedpassword, "none")==0)
    temppw[0]=0;
  else
    uncryptpassword(cryptedpassword, temppw);

  dnstring(temppw); // passwords are case-insensitive

  std::string md5_data;
  md5_data+=up.name;
  md5_data+=temppw;
  char md5buf[16];
  md5_calc((unsigned char *)md5buf,
           (const unsigned char *)md5_data.c_str(),
           md5_data.size()+1);

  std::string bufcoded;
  base64_encode(std::string(md5buf, sizeof(md5buf)), bufcoded);

  out << up.name << "=" << bufcoded;

  out << std::endl;
 }

/*******************************************************************\

Function: auth

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void do_auth(std::ostream &out)
 {
  userlistt userlist;

  get_userlist(userlist);

  for(userlistt::const_iterator it=userlist.begin();
      it!=userlist.end(); it++)
    do_auth(out, *it);
 }

