/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <dblib/os_dep_file.h>
#include <libdbox/userdb.h>
#include <libdbox/config.h>

#include "mail.h"
#include "userlist.h"

/*******************************************************************\

Function: parse_domains

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void parse_domains(const std::string &domains,
                   std::list<std::string> &domain_list)
 {
  const char *tptr, *tptr2;

  tptr=domains.c_str();
  while(*tptr!=0)
   {
    tptr2=tptr;
    while(*tptr2!=0 && *tptr2!=' ' && *tptr2!='\t' && 
          *tptr2!=';' && *tptr2!=',')
      tptr2++;

    if(*tptr=='.') tptr++;

    domain_list.push_back(std::string(tptr, tptr2-tptr));

    tptr=tptr2;
    if(*tptr!=0)
      while(*tptr==' ' || *tptr=='\t' ||
            *tptr==';' || *tptr==',')
        tptr++;
   }
 }

/*******************************************************************\

Function: mail

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void do_mail(std::ostream &out, const std::string &user)
 {
  userpt up(user.c_str());

  if(up.read())
   {
    std::cerr << "failed to read profile of user "
              << user << std::endl;
    return;
   }

  userpt up2;
  grouppt gp;

  get_profiles(up, up2, gp);

  std::string alias, domains;
  up.get("alias", alias);
  up.get("domains", domains);

  bool bouncemail;

   {
    char data[TEXTLEN];

    up2.get("bouncemail", data);
    if(data[0]==0)
      gp.get("bouncemail", data);

    bouncemail=atoi(data);
   }

  //if(bouncemail) return;

  //bool is_system=atoi(up|"system");

  //if(is_system) std::cout << "SYSTEM: " << up.name << std::endl;

  std::string dest=up2.name;

  if(dest==config.boxname) dest="localhost";

  if(bouncemail)
    dest="";

  out << up.name << "@localhost" << "=" << dest << std::endl;

  if(domains!="" && !bouncemail)
   {
    std::list<std::string> domain_list;

    parse_domains(domains, domain_list);

    for(std::list<std::string>::const_iterator it=domain_list.begin();
        it!=domain_list.end(); it++)
     {
      out << up.name << "." << *it
          << "=" << dest << std::endl;
     }
   }
 }

/*******************************************************************\

Function: mail

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void do_mail(std::ostream &out)
 {
  userlistt userlist;

  get_userlist(userlist);

  for(userlistt::const_iterator it=userlist.begin();
      it!=userlist.end(); it++)
    do_mail(out, *it);
 }

