/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

//#include <folder.h>
#include <stringtools.h>
#include <dblib/sbase64.h>
#include <dbclient/logstr.h>
#include <dbclient/client.h>

#ifndef DBOX_STANDALONE
#include <dboxversion.h>
#endif

#include "dbimap4d.h"
#include "imap4fetch.h"
#include "imap4store.h"
#include "imap4copy.h"

/*******************************************************************\

Function: dbimap4dt::docapability

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::docapability()
 {
  out << "* " << command
      << " IMAP4 IMAP4rev1 AUTH=LOGIN ACL NAMESPACE\r\n";
 }

/*******************************************************************\

Function: dbimap4dt::do_command_authenticated

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::do_command_authenticated()
 {
  if(command=="CAPABILITY")
    docapability();
  else if(command=="NOOP")
   {
   }
  else if(command=="DK")
   {
    for(unsigned argnr=0; argnr<arguments.size(); argnr++)
     {
      out << "DK: \"" << arguments[argnr].data << "\" "
          << "end_list: " << arguments[argnr].end_list << " "
          << "begin_list: " << arguments[argnr].begin_list
          << "\r\n";
     }
   }
  else if(command=="LOGOUT")
   {
    out << "* BYE as requested\r\n";

    logstr << user << ": logout by user request";
    logstr.log("dbimap4dt::do_command_authenticated");

    logout=TRUE;
   }
  else if(command=="AUTHENTICATE" ||
          command=="LOGIN")
    error("Sorry, already authenticated");
  else if(command=="SELECT" ||
          command=="EXAMINE")
    doselect();
  else if(command=="CREATE")
    docreate();
  else if(command=="NAMESPACE")
    donamespace();
  else if(command=="GETACL")
    dogetacl();
  else if(command=="LISTRIGHTS")
    dolistrights();
  else if(command=="MYRIGHTS")
    domyrights();
  else if(command=="SETACL")
    dosetacl();
  else if(command=="DELETEACL")
    dodeleteacl();
  else if(command=="DELETE")
    dodelete();
  else if(command=="RENAME")
    dorename();
  else if(command=="SUBSCRIBE")
    dosubscribe(TRUE);
  else if(command=="UNSUBSCRIBE")
    dosubscribe(false);
  else if(command=="LIST" || command=="LSUB")
    dolsublist();
  else if(command=="APPEND")
    doappend();
  else if(command=="CHECK")
   {
   }
  else if(command=="CLOSE")
   {
    if(!mailbox_selected)
      error("no mailbox selected");
    else
     {
      // TODO: remove deleted messages
      close();
      mailbox_selected=false;
     }
   }
  else if(command=="EXPUNGE")
   {
    if(!mailbox_selected)
      error("no mailbox selected");
    else
      doexpunge();
   }
  else if(command=="SEARCH")
   {
    if(!mailbox_selected)
      error("no mailbox selected");
    else
      dosearch(false);
   }
  else if(command=="FETCH")
   {
    if(!mailbox_selected)
      error("no mailbox selected");
    else
     {
      fetch_set_commandt set_command(*this);
      set_command.do_set_command(false);
     }
   }
  else if(command=="PARTIAL")
   {
    error("command is obsolete");
   }
  else if(command=="STORE")
   {
    if(!mailbox_selected)
      error("no mailbox selected");
    else
     {
      store_set_commandt set_command(*this);
      set_command.do_set_command(false);
     }
   }
  else if(command=="COPY")
   {
    if(!mailbox_selected)
      error("no mailbox selected");
    else
     {
      copy_set_commandt set_command(*this);
      set_command.do_set_command(false);
     }
   }
  else if(command=="STATUS")
   {
    dostatus();
   }
  else if(command=="UID")
   {
    if(!mailbox_selected)
      error("no mailbox selected");
    else
     {
      upstring(arguments[0].data);

      if(arguments[0].data=="FETCH")
       {
        fetch_set_commandt set_command(*this);
        set_command.do_set_command(TRUE);
       }
      else if(arguments[0].data=="COPY")
       {
        copy_set_commandt set_command(*this);
        set_command.do_set_command(TRUE);
       }
      else if(arguments[0].data=="STORE")
       {
        store_set_commandt set_command(*this);
        set_command.do_set_command(TRUE);
       }
      else if(arguments[0].data=="SEARCH")
       {
        dosearch(TRUE);
       }
      else
        error("not implemented");

      if(command_ok)
       {
        status_update();
        out << tag << " OK " << command << " "
            << arguments[0].data << " completed\r\n";
        command_ok=false;
       }
     }
   }
  else
   {
    out << tag << " BAD command unknown\r\n";
    command_ok=false;
   }
 }

/*******************************************************************\

Function: dbimap4dt::do_command_not_authenticated

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::do_command_not_authenticated()
 {
  if(command=="CAPABILITY")
    docapability();
  else if(command=="NOOP")
   {
   }
  else if(command=="DK")
   {
    for(unsigned argnr=0; argnr<arguments.size(); argnr++)
     {
      out << "DK: \"" << arguments[argnr].data << "\" "
          << "end_list: " << arguments[argnr].end_list << " "
          << "begin_list: " << arguments[argnr].begin_list
          << "\r\n";
     }
   }
  else if(command=="LOGOUT")
   {
    out << "* BYE as requested\r\n";

    logstr << "logout by user request";
    logstr.log("dbimap4dt::do_command_not_authenticated");

    logout=TRUE;
   }
  else if(command=="AUTHENTICATE")
   {
    if(arguments.size()==0)
      error("Sorry, AUTHENTICATE needs arguments");
    else
     {
      std::string method=arguments[0].data;
      upstring(method);

      if(method=="LOGIN")
       {
        out << "+ VXNlciBOYW1lAA==\r\n";
        if(getcommand()) { channel_lost(); return; }
        user=sbase64_decode(commandline);

        out << "+ UGFzc3dvcmQA\r\n";
        if(getcommand()) { channel_lost(); return; }

        if(authenticate(sbase64_decode(commandline)))
          error("invalid password");
       }
      else
        error("Sorry, authentication method "+
              arguments[0].data+" is not supported "
              "up to now");
     }
   }
  else if(command=="LOGIN")
   {
    if(arguments.size()!=2)
     {
      out << tag << " BAD invalid username\r\n";
      command_ok=false;
     }
    else
     {
      user=arguments[0].data;
      if(authenticate(arguments[1].data))
        error("invalid password");
     }
   }
  else
   {
    out << tag << " BAD command unknown\r\n";
    command_ok=false;
   }
 }

/*******************************************************************\

Function: dbimap4t::dbimap4d

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::dbimap4d()
 {
  if(client.config["maildir"]=="")
   {
    out << "* BYE Mail system configuration error\r\n";
    return;
   }

  out << "* OK " << client.config["hostname"]
      << " DBIMAP Server (version "
      << DBOXVERSIONNR << ") ready.\r\n";

  try
   {
    getpeer();

    logout=false;
    last_status_update=0;
    last_index_file_size=0;
    largest_id=0;
    mailbox_selected=false;

    if(preauth_user!="")
     {
      user=preauth_user;
      authenticated=true;
     }

    while(!logout)
     {
      if(get_and_parse_command())
       {
        channel_lost();
        return;
       }

      command_ok=TRUE;

      if(authenticated)
        do_command_authenticated();
      else
        do_command_not_authenticated();

      if(command_ok)
       {
        status_update();
        out << tag << " OK " << command << " completed\r\n";
       }
   
      fflush(stdout);
     }
   }

  catch(const excpt &e)
   {
    error(e.what());
    out << "* BYE giving up\r\n";
   }
 }

/*******************************************************************\

Function: dbimap4d

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4d(const std::string &preauth)
 {
  dbimap4dt dbimap4d(std::cin, std::cout);

  dbimap4d.preauth_user=preauth;

  dbimap4d.dbimap4d();
 }
