/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <memory>
#include <string.h>
#include <iostream>
#include <fstream>

#include <sfolderrfc.h>
#include <stringtools.h>
#include <dblib/i2string.h>
#include <dblib/os_dep_file.h>
#include <dbsmtpd/folder_insert.h>
#include <dbsmtpd/folder_profile.h>

#include "dbimap4d.h"

/*******************************************************************\

Function: dbimap4dt::doappend2

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::doappend2(mailfoldert &new_mailfolder)
 {
  new_mailfolder.lock("dbimap4dt::doappend2");

  dword nr;

  folder_insert(rfcm, new_mailfolder.get_path(),
                false, "", "", nr);

  folderindext fi;

  std::string filename=
    build_path(new_mailfolder.get_path(), "index");

  std::fstream index(filename.c_str(),
                     std::ios::in|std::ios::out|std::ios::binary);

  if(!index)
   {
    error("internal error 1");
    return;
   }

  index.seekp(-(signed)sizeof(fi), std::ios::end);
  std::streampos pos=index.tellp();
  if(!index.read((char *)&fi, sizeof(fi)))
   {
    error("internal error 2 "+i2string((unsigned long)pos));
    return;
   }

  unsigned argnr=1;
  if(arguments[argnr].begin_list) // Flags
   {
    for(argnr++; argnr<arguments.size() && 
        !arguments[argnr].end_list; argnr++)
     {
      const char *item=arguments[argnr].data.c_str();
      if(stricmp(item, "\\Answered")==0)
        fi.flags.imapflags|=DBOXIMAP4_FLAG_Answered;
      else if(stricmp(item, "\\Flagged")==0)
        fi.flags.imapflags|=DBOXIMAP4_FLAG_Flagged;
      else if(stricmp(item, "\\Deleted")==0)
        fi.flags.deleted=TRUE;
      else if(stricmp(item, "\\Seen")==0)
        fi.flags.seen=TRUE;
      else if(stricmp(item, "\\Draft")==0)
        fi.flags.imapflags|=DBOXIMAP4_FLAG_Draft;
     }
   }

  if(!arguments[argnr].literal) // internaldate
    fi.received=parse_date(arguments[argnr].data);

  index.seekp(pos, std::ios::beg);
  index.write((char *)&fi, sizeof(fi));

  new_mailfolder.unlock();

  // APPENDUID: RFC2359
  folder_profilet new_fp(new_mailfolder);
  dword uidvalidity=new_fp.get_int("imap.uidvalidity");
  if(uidvalidity==0) uidvalidity=1;

  out << tag << " OK [APPENDUID " << uidvalidity << " " 
      << fi.id << "] APPEND completed\r\n";
 }

/*******************************************************************\

Function: dbimap4dt::doappend

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::doappend()
 {
  unsigned int args=arguments.size();

  command_ok=false;

  if(args<2)
   {
    error("give arguments");
    return; 
   }

  std::string mailbox_name=arguments[0].data;
  mailfoldert new_mailfolder;

  if(process_mailbox_name(mailbox_name, new_mailfolder))
    return;

  folder_grantst grants;
  get_folder_grants(new_mailfolder, grants);
  if(!grants.insert)
   {
    error("permission denied");
    return;
   }

   {
    size_t size=arguments[args-1].data.size();

    std::auto_ptr<char> message(new char[size]);

    rfcm.data=&(*message);
    memcpy((char *)rfcm.data, arguments[args-1].data.data(), size);
    rfcm.endoffile=rfcm.data+size;
    rfcm.endofmail=rfcm.endoffile;

    if(rfcm.parse() || rfcm.headerlen==0)
      error("invalid message");
    else
      doappend2(new_mailfolder);
   }
 }

