/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: COPY command

*/

#include <iostream>
#include <fstream>

#include <dblib/i2string.h>
#include <dblib/os_dep_file.h>
#include <dbsmtpd/folder_insert.h>
#include <dbsmtpd/folder_profile.h>

#include "imap4copy.h"

/*******************************************************************\

Function: copy_set_commandt::start

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool copy_set_commandt::start()
 {
  std::string new_mailbox_name=dbimap4d.arguments[argnr].data, 
              new_folder_owner, new_subfolder, new_folder_name;

  if(dbimap4d.process_mailbox_name(new_mailbox_name, new_mailfolder))
    return true;

  folder_grantst new_grants;
  dbimap4d.get_folder_grants(new_mailfolder, new_grants);

  if(!new_grants.insert)
   {
    dbimap4d.out << dbimap4d.tag << " NO permission denied\r\n";
    dbimap4d.command_ok=false;
    return true;
   }

  new_mailfolder.lock("copy_set_commandt::start");

  return false;
 }

/*******************************************************************\

Function: copy_set_commandt::msg

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool copy_set_commandt::msg()
 {
  if(dbimap4d.get_body(nr-1))
   {
    dbimap4d.out << "* " << nr << " ERROR: failed to get message\r\n";
    return false;
   }

  if(dbimap4d.rfcm.parse() || 
     dbimap4d.rfcm.headerlen==0)
   {
    dbimap4d.out << "* " << nr << " ERROR: invalid message\r\n";
    return false;
   }

  dword new_nr;

  folder_insert(dbimap4d.rfcm,
                new_mailfolder.get_path(),
                false, // forceextern
                "", // envelope_to
                "", // envelope_from
                new_nr);

  folderindext new_fi;

  std::string filename=build_path(new_mailfolder.get_path(), "index");
  std::fstream new_index(filename.c_str(), std::ios::in|std::ios::out|std::ios::binary);

  if(!new_index)
   {
    dbimap4d.out << dbimap4d.tag << " NO internal error 1\r\n";
    dbimap4d.command_ok=false;
    return true;
   }

  new_index.seekp(-(signed)sizeof(new_fi), std::ios::end);
  std::streampos pos=new_index.tellp();
  if(!new_index.read((char *)&new_fi, sizeof(new_fi)))
   {
    dbimap4d.out << dbimap4d.tag << " NO internal error 2 " << pos << "\r\n";
    dbimap4d.command_ok=false;
    return true;
   }

  new_fi.flags.imapflags=dbimap4d.fi.flags.imapflags;
  new_fi.flags.deleted=dbimap4d.fi.flags.deleted;
  new_fi.flags.seen=dbimap4d.fi.flags.seen;
  new_fi.received=dbimap4d.fi.received;

  new_index.seekp(pos, std::ios::beg);
  new_index.write((char *)&new_fi, sizeof(new_fi));

  if(copyuid!="") copyuid+=",";
  copyuid+=i2string(unsigned(new_fi.id));

  return false;
 }

/*******************************************************************\

Function: copy_set_commandt::end

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void copy_set_commandt::end()
 {
  // COPYUID: RFC2359

  folder_profilet new_fp(new_mailfolder);
  
  dword uidvalidity=new_fp.get_int("imap.uidvalidity");
  if(uidvalidity==0) uidvalidity=1;

  new_mailfolder.unlock();

  dbimap4d.out << dbimap4d.tag << " OK [COPYUID " << uidvalidity << " " 
               << copyuid << "] COPY completed\r\n";

  dbimap4d.command_ok=false;
 }
