/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <fstream>

#include <stringtools.h>
#include <dblib/os_dep_file.h>
#include <dblib/i2string.h>
#include <dbclient/logstr.h>
#include <dbsmtpd/folder_profile.h>

#include "mod-utf7.h"
#include "dbimap4d.h"

/*******************************************************************\

Function: dbimap4dt::docreate

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::docreate()
 {
  if(arguments.size()!=1)
   {
    out << tag << " NO give a mailbox name\r\n";
    command_ok=FALSE;
    return;
   }

  std::string mailbox_name=decode_mod_utf7(arguments[0].data);

  if(mailbox_name=="")
   {
    out << tag << " NO give a non-empty mailbox name\r\n";
    command_ok=FALSE;
    return;
   }
  
  if(mailbox_name[mailbox_name.size()-1]=='/')
   {
    out << tag << " OK ignored\r\n";
    command_ok=FALSE;
    return;
   }

  if(stringcasecmp(mailbox_name, "INBOX")==0)
   {
    out << tag << " NO mailbox already exists\r\n";
    command_ok=FALSE;
    return;
   }

  if(check_mailbox_name(mailbox_name))
   {
    out << tag << " NO mailbox name is not valid\r\n";
    command_ok=FALSE;
    return;
   }

  std::string subfolder,
              user_path=get_user_path();

   {
    lockt folders_lock("folders", user, "dbimap4dt::docreate");

    get_folder_list();

    if(find_folder(mailbox_name)!=folders.end())
     {
      error("mailbox already exists");
      return;
     }

    unsigned i;
    for(i=0; i<folders.size(); i++)
      if(folders[i].name=="")
        break;

     {
      folder_et folder_e;

      folder_e.name=mailbox_name;
      folder_e.subscribed=false; // not automatically subscribed to

      if(i==folders.size())
        folders.push_back(folder_e);
      else
        folders[i]=folder_e;
     }

    subfolder=i2string(i);

     {
      std::string outfilename=build_path(user_path, "folders");
      std::ofstream outfile(outfilename.c_str(), std::ios::out);

      for(i=0; i<folders.size(); i++)
        outfile << folders[i].name << "\n";
     }

    // folder lock gets removed here
   }

   {
    std::string foldersdir=build_path(user_path, "mfolders");
    mkdir(foldersdir.c_str(), 0700);
    std::string subfolderdir=build_path(foldersdir, subfolder);
    mkdir(subfolderdir.c_str(), 0700);
   }

   {
    mailfoldert new_mailfolder(user, subfolder);

    new_mailfolder.lock("dbimap4dt::docreate");
    folder_profilet new_fp(new_mailfolder);
    
    new_fp.set("imap.uidvalidity", time(NULL));
    new_fp.set("nextid", 1);

    if(new_fp.write())
     {
      logstr << "failed to write " << new_fp.get_filename();
      logstr.log("dbimap4dt::docreate");
      error("internal error");
     }
   }
 }
