/*

  DBOX Module
  Copyright (C) 1994-2003 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <iostream>
#include <sstream>

#include <stringtools.h>
#include <dblib/os_dep_file.h>
#include "imap4fetch.h"

/*******************************************************************\

Function: fetch_set_commandt::count_lines

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

unsigned fetch_set_commandt::count_lines(const char *msg, size_t size)
 {
  unsigned lines=0;

  for(; size>0; size--, msg++)
    if(*msg=='\n') lines++;

  return lines;
 }

/*******************************************************************\

Function: fetch_set_commandt::count_lines

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

unsigned fetch_set_commandt::count_lines(mimestructt &mimestruct)
 {
  return count_lines(dbimap4d.rfcm.data+
    mimestruct.offset+mimestruct.headerlen,
    mimestruct.len);
 }

/*******************************************************************\

Function: fetch_set_commandt::filter_header

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::filter_header(int startargnr)
 {
  const char *id, *text, *start;
  std::ostringstream header;

  if(dbimap4d.get_header(nr-1))
    return;

  start=dbimap4d.rfcm.data;

  while(!dbimap4d.rfcm.parseline(&start, &id, &text) && id!=NULL)
   {
    int filterargnr;
    std::string text;

    for(filterargnr=startargnr;
        !dbimap4d.arguments[filterargnr].end_list;
        filterargnr++)
     {
      dbimap4d.rfcm.copyheaderid(text, id);
      if(stringcasecmp(text, dbimap4d.arguments[filterargnr].data)==0)
       {
        header.write(id, start-id);        
        break;
       }
     }
   }

  header << "\r\n";

  dbimap4d.out << '{' << header.tellp() << "}\r\n";

  dbimap4d.out.write(header.str().c_str(), header.str().size());
 }

/*******************************************************************\

Function: fetch_set_commandt::fetch_body_fields

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::fetch_body_fields(mimestructt &mimestruct)
 {
  unsigned i;

  dbimap4d.list_item(); // body_fld_param
  dbimap4d.begin_list();
  for(i=0; i<mimestruct.content_type.attributes.size(); i++)
   {
    dbimap4d.list_item();
    dbimap4d.send_string(mimestruct.content_type.attributes[i].name);
    dbimap4d.list_item();
    dbimap4d.send_string(mimestruct.content_type.attributes[i].value);
   }
  dbimap4d.end_list();

  dbimap4d.list_item(); // body_fld_id
  dbimap4d.out << "NIL";

  dbimap4d.list_item(); // body_fld_desc
  dbimap4d.out << "NIL";

  dbimap4d.list_item(); // body_fld_enc
  if(mimestruct.content_transfer_encoding.value!="")
    dbimap4d.send_string(mimestruct.content_transfer_encoding.value);
  else
    dbimap4d.send_string("7BIT");

  dbimap4d.list_item(); // body_fld_octets
  dbimap4d.out << mimestruct.len;
 }

/*******************************************************************\

Function: fetch_set_commandt::fetch_bodystructure

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::fetch_bodystructure(mimestructt &mimestruct)
 {
  dbimap4d.begin_list();

  if(stringcasecmp(mimestruct.content_type.value, "multipart")==0) // mpart
   {
    if(mimestruct.sub.size()==0)
     {
      dbimap4d.list_item();
      dbimap4d.out << "NIL";
     }
    else
     {
      dbimap4d.list_item();

      for(unsigned i=0; i<mimestruct.sub.size(); i++)
        fetch_bodystructure(mimestruct.sub[i]);

      dbimap4d.list_item();
      dbimap4d.send_string(mimestruct.content_type.subvalue);
     }
   }
  else
   {
    dbimap4d.list_item();

    dbimap4d.send_string(mimestruct.content_type.value);

    if(stringcasecmp(mimestruct.content_type.value, "text")==0) // basic
     {
      dbimap4d.list_item(); dbimap4d.send_string(mimestruct.content_type.subvalue);
      fetch_body_fields(mimestruct);
      dbimap4d.list_item(); dbimap4d.out << count_lines(mimestruct);
     }
    else if(stringcasecmp(mimestruct.content_type.value, "message")==0 &&
            stringcasecmp(mimestruct.content_type.subvalue, "rfc822")==0) // msg
     {
      rfcmessaget subpartm;

      subpartm.data=dbimap4d.rfcm.data+mimestruct.offset+mimestruct.headerlen;
      subpartm.endofmail=subpartm.endoffile=subpartm.data+mimestruct.len;

      mimestructt substruct;
      substruct.offset=mimestruct.offset+mimestruct.headerlen;
      substruct.mimeparse(subpartm);

      dbimap4d.list_item(); dbimap4d.send_string(mimestruct.content_type.subvalue);
      fetch_body_fields(mimestruct);
      dbimap4d.list_item(); fetch_envelope(subpartm);
      dbimap4d.list_item(); fetch_bodystructure(substruct);
      dbimap4d.list_item(); dbimap4d.out << count_lines(mimestruct);
     }
    else // basic
     {
      dbimap4d.list_item();
      dbimap4d.send_string(mimestruct.content_type.subvalue);
      fetch_body_fields(mimestruct);
     }
   }
  
  dbimap4d.end_list();
 }

/*******************************************************************\

Function: fetch_set_commandt::fetch_bodystructure

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::fetch_bodystructure(const std::string &what)
 {
  if(dbimap4d.get_structure(nr-1))
    return;

  dbimap4d.list_item();
  dbimap4d.out << what;
  dbimap4d.list_item();
  fetch_bodystructure(*dbimap4d.mimestruct);
 }

/*******************************************************************\

Function: fetch_set_commandt::fetch_body_parse

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::fetch_body_parse(const std::string &what,
  bool &peek, std::string &specification, std::string &part,
  bool &partial, dword &partial_start, dword &partial_size)
 {
  partial=peek=FALSE;
  const char *what_ptr=what.c_str();

  partial_start=0;
  partial_size=(dword)-1;

  if(strnicmp(what_ptr, "BODY", 4)==0)
    what_ptr+=4;

  if(strnicmp(what_ptr, ".PEEK", 5)==0)
   { what_ptr+=5; peek=TRUE; }

  if(what_ptr[0]=='[')
   {
    what_ptr++;

    char _part[MAXARGLEN];
    strmaxcpy(_part, what_ptr, MAXARGLEN-1);
    char *tptr=_part;

    while(isdigit(tptr[0]) || tptr[0]=='.') tptr++;

    char _specification[MAXARGLEN];
    strmaxcpy(_specification, tptr, MAXARGLEN-1);
    *tptr=0;

    tptr=strchr(_specification, ']');
    if(tptr!=NULL) *(tptr++)=0;

    part=_part;
    specification=_specification;
   }

  // TODO: case header.fields: next argument

  what_ptr=strchr(what.c_str(), '<');
  if(what_ptr!=NULL)
   {
    partial=TRUE;
    what_ptr++;
    partial_start=atol(what_ptr);
    while(isdigit(*what_ptr)) what_ptr++;
    if(*what_ptr=='.') 
     {
      what_ptr++;
      partial_size=atol(what_ptr);
     }
   } 
 }

/*******************************************************************\

Function: fetch_set_commandt::fetch_body4

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::fetch_body4(const char *ptr,
  dword data_size, bool partial,
  dword partial_start, dword partial_size)
 {
  #if 0
  // Netscape Bug Workaround
  if(partial && partial_start==0)
    partial=FALSE;
  #endif

  if(partial)
    dbimap4d.out << "<" << partial_start << ">";

  dbimap4d.list_item();

  if(partial)
   {
    if(partial_start>=data_size)
      data_size=0;
    else
     {
      data_size-=partial_start;
      if(data_size>partial_size) data_size=partial_size;
     }
   }
  else
    partial_start=0;

  dbimap4d.out << '{' << data_size << "}\r\n";

  if(data_size!=0)
    dbimap4d.out.write(ptr+partial_start, data_size);
 }

/*******************************************************************\

Function: fetch_set_commandt::fetch_body3

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::fetch_body3(
  const std::string &specification,
  bool partial, dword partial_start, dword partial_size,
  const std::string &display_response)
 {
  if(specification=="")
   {
    if(!dbimap4d.get_body(nr-1))
     {
      dbimap4d.list_item();
      dbimap4d.out << display_response;

      fetch_body4(dbimap4d.rfcm.data,
                  dbimap4d.rfcm.headerlen+dbimap4d.rfcm.len,
                  partial, partial_start, partial_size);
     }
   }
  else if(stricmp(specification.c_str(), "HEADER")==0)
   {
    if(!dbimap4d.get_header(nr-1))
     {
      dbimap4d.list_item();
      dbimap4d.out << display_response;
      fetch_body4(dbimap4d.rfcm.data,
                  dbimap4d.rfcm.headerlen,
                  partial, partial_start, partial_size);
     }
   }
  else if(stricmp(specification.c_str(), "TEXT")==0)
   {
    if(!dbimap4d.get_body(nr-1))
     {
      dbimap4d.list_item();
      dbimap4d.out << display_response;
      fetch_body4(dbimap4d.rfcm.data+dbimap4d.rfcm.headerlen,
                  dbimap4d.rfcm.len,
                  partial, partial_start, partial_size);
     }
   }
  else if(strnicmp(specification.c_str(), "HEADER.FIELDS", 13)==0)
   {
    if(!dbimap4d.get_header(nr-1))
     {
      bool inverse=FALSE;
      int startargnr;

      if(stricmp(specification.c_str()+13, ".NOT")==0)
        inverse=TRUE;

      argnr++;
      if(!dbimap4d.arguments[argnr].begin_list) // parse error
        return;

      argnr++;
      startargnr=argnr;

      dbimap4d.list_item();
      dbimap4d.out << "BODY[" << specification;
      dbimap4d.list_item();
      dbimap4d.begin_list();

      while(argnr<dbimap4d.arguments.size() && 
            !dbimap4d.arguments[argnr].end_list)
       {
        dbimap4d.list_item();
        //upstring(arguments[argnr].data); // Hm?
        dbimap4d.send_string(dbimap4d.arguments[argnr].data);
        argnr++;
       }

      dbimap4d.end_list();

      dbimap4d.out << ']';
      if(dbimap4d.arguments[argnr].data=="]") argnr++;

      dbimap4d.list_item();

      filter_header(startargnr);
     }
   }
 } 

/*******************************************************************\

Function: fetch_set_commandt::fetch_body3_part2

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::fetch_body3_part2(
  mimestructt &mimestruct,
  const std::string &specification,
  bool partial, dword partial_start, dword partial_size,
  const std::string &display_response)
 {
  if(specification=="")
   {
    dbimap4d.list_item();
    dbimap4d.out << display_response;
    fetch_body4(dbimap4d.rfcm.data+
                mimestruct.offset+mimestruct.headerlen,
                mimestruct.len,
                partial, partial_start, partial_size);
   }
  else if(stricmp(specification.c_str(), "MIME")==0)
   {
    dbimap4d.list_item();
    dbimap4d.out << display_response;
    fetch_body4(dbimap4d.rfcm.data+
                mimestruct.offset, 
                mimestruct.headerlen,
                partial, partial_start, partial_size);
   }
  else
   {
    if(mimestruct.sub.size()==1)
     {
      struct mimestructt &_mimestruct=mimestruct.sub[0];

      if(stricmp(specification.c_str(), "HEADER")==0)
       {
        dbimap4d.list_item();
        dbimap4d.out << display_response;
        fetch_body4(dbimap4d.rfcm.data+
                    _mimestruct.offset, 
                    _mimestruct.headerlen,
                    partial, partial_start, partial_size);
       }
      else if(stricmp(specification.c_str(), "TEXT")==0)
       {
        dbimap4d.list_item();
        dbimap4d.out << display_response;
        fetch_body4(dbimap4d.rfcm.data+
                    _mimestruct.offset+mimestruct.headerlen,
                    _mimestruct.len,
                    partial, partial_start, partial_size);
       }
      else if(strnicmp(specification.c_str(), "HEADER.FIELDS", 13)==0)
       { // TODO
        bool inverse=FALSE;
        int startargnr;

        if(stricmp(specification.c_str()+13, ".NOT")==0)
          inverse=TRUE;

        argnr++;
        if(!dbimap4d.arguments[argnr].begin_list) // parse error
          return;

        argnr++;
        startargnr=argnr;

        dbimap4d.list_item();
        dbimap4d.out << "BODY[" << specification;
        dbimap4d.list_item();
        dbimap4d.begin_list();

        while(argnr<dbimap4d.arguments.size() && 
              !dbimap4d.arguments[argnr].end_list)
         {
          dbimap4d.list_item();
          //upstring(arguments[argnr].data); // Hm?
          dbimap4d.send_string(dbimap4d.arguments[argnr].data);
          argnr++;
         }

        dbimap4d.end_list();

        dbimap4d.out << ']';
        if(dbimap4d.arguments[argnr].data=="]") argnr++;

        dbimap4d.list_item();

        filter_header(startargnr);
       }
     }
   }
 }

/*******************************************************************\

Function: fetch_set_commandt::fetch_body3_part

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::fetch_body3_part(
  mimestructt &mimestruct,
  const std::string &specification, const std::string &part,
  bool partial, dword partial_start, dword partial_size,
  const std::string &display_response)
 {
  if(part=="1" && mimestruct.sub.size()==0 &&
     specification=="")
   {
    fetch_body3("TEXT", partial,
      partial_start, partial_size,
      display_response);
   }
  else
   {
    unsigned sub=atoi(part.c_str());

    if(!isdigit(part[0]) ||
       sub>mimestruct.sub.size() || 
       sub==0)
     {
      dbimap4d.list_item(); 
      dbimap4d.out << display_response;
      dbimap4d.list_item();
      dbimap4d.out << "NIL";
     }
    else
     {
      const char *tptr=part.c_str();
      while(isdigit(*tptr)) tptr++;
      if((*tptr)=='.') tptr++;

      if(tptr[0]==0)
       {
        fetch_body3_part2(mimestruct.sub[sub-1],
          specification,
          partial, partial_start, partial_size,
          display_response);
       }
      else
       {
        fetch_body3_part(mimestruct.sub[sub-1], specification,
          tptr, partial, partial_start, partial_size, display_response);
       }
     }
   }
 } 

/*******************************************************************\

Function: fetch_set_commandt::fetch_body2

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::fetch_body2(const std::string &what,
                                     const std::string &response)
 {
  bool peek, partial;
  std::string specification, part;
  dword partial_start, partial_size;

  fetch_body_parse(what, peek, specification, part,
                   partial, partial_start, partial_size);

  std::string display_response;
  if(response=="")
    display_response="BODY["+part+specification+"]";
  else
    display_response=response;

  if(part=="")
   {
    fetch_body3(specification, partial,
      partial_start, partial_size, display_response);
   }
  else
   {
    if(dbimap4d.get_structure(nr-1))
      return;

    fetch_body3_part(*dbimap4d.mimestruct,
      specification, part, partial,
      partial_start, partial_size, display_response);
   }

  if(!peek)
   {
    dbimap4d.get_folderindex(nr-1);
    if(!dbimap4d.fi.flags.seen)
     {
      dbimap4d.mailfolder.lock("fetch_set_commandt::fetch_body2");

      lseek(dbimap4d.index, (nr-1)*sizeof dbimap4d.fi, SEEK_SET);
      read(dbimap4d.index, &dbimap4d.fi, sizeof dbimap4d.fi);
      dbimap4d.fi.flags.seen=true;
      lseek(dbimap4d.index, (nr-1)*sizeof dbimap4d.fi, SEEK_SET);
      write(dbimap4d.index, &dbimap4d.fi, sizeof dbimap4d.fi);

      dbimap4d.mailfolder.unlock();

      fetch("FLAGS");
     }
   }
 }

/*******************************************************************\

Function: trim

  Inputs:

 Outputs:

 Purpose: 

\*******************************************************************/

static void trim(std::string &str)
 {
  const char *tptr, *tptr2;

  tptr2=tptr=str.c_str();
  while(isspace(*tptr)) tptr++;

  if(tptr!=tptr2)
   {
    std::string tmp(tptr);
    str.swap(tmp);
   }

  while(str.size()!=0 && isspace(str[str.size()-1]))
    str.resize(str.size()-1);
 }

/*******************************************************************\

Function: fetch_set_commandt::address_to_imap

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::address_to_imap(std::ostringstream &realname,
                                         std::ostringstream &email)
 {
  dbimap4d.begin_list();

   {
    std::string str(realname.str());
    trim(str);
    dbimap4d.list_item(); dbimap4d.send_string(str);
   }

  dbimap4d.list_item(); dbimap4d.out << "NIL";

   {
    std::string str=email.str();
    trim(str);

    char *temp=strdup(str.c_str());
    
    char *tptr=strchr(temp, '@');
    if(tptr!=NULL) *(tptr++)=0;
    dbimap4d.list_item(); dbimap4d.send_string(str);
    dbimap4d.list_item(); dbimap4d.send_string(tptr==NULL?"":tptr);
    delete temp;
   }

  dbimap4d.end_list();
 }

/*******************************************************************\

Function: fetch_set_commandt::address_to_imap

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::address_to_imap(const char *&address)
 {
  int state=0;
  bool done=FALSE;
  std::ostringstream text1, text2;

  while(!done)
   {
    if(address[0]=='"')
     {
      address++;
      while(address[0]!='"' && address[0]!=0)
        text1 << (address++)[0];
     }
    else if(state==0)
     {
      switch(address[0])
       {
       case '<':  // address has form realname <user@host>
        state=1;
        break;

       case '(':  // address has form user@host (realname)
        state=2;
        break;

       case ',':  // address has form user@host
       case 0:
        address_to_imap(text2, text1);
        done=TRUE;
        break;

       default:
        text1 << address[0];
       }
     }
    else if(state==1) // address has form realname <user@host>
     {
      if(address[0]=='>' || address[0]==0)
       {
        address_to_imap(text1, text2);
        done=TRUE;
       }
      else
        text2 << address[0];
     }
    else if(state==2) // address has form user@host (realname)
     {
      if(address[0]==')' || address[0]==0)
       {
        address_to_imap(text2, text1);
        done=TRUE;
       }
      else 
        text2 << address[0];
     }

    if(address[0]!=0) address++;
   }

  while(isspace(address[0]) || address[0]==',') address++;
 }

/*******************************************************************\

Function: fetch_set_commandt::address_list_to_imap

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::address_list_to_imap(const std::string &address)
 {
  if(address=="")
   {
    dbimap4d.out << "NIL";
    return;
   }

  const char *tptr=address.c_str();

  dbimap4d.begin_list();
  while(tptr[0]!=0)
   {
    dbimap4d.list_item();
    address_to_imap(tptr);
   }
  dbimap4d.end_list();
 }

/*******************************************************************\

Function: fetch_set_commandt::send_header

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::send_header(rfcmessaget &rfcm,
                                     const std::string &name)
 {
  std::string header; 

  if(rfcm.findsingleheader(header, name))
   {
    if(stringcasecmp(name, "reply-to")==0 || // default those fields
       stringcasecmp(name, "sender")==0)
      send_header(rfcm, "from");
    else
      dbimap4d.out << "NIL";
   }
  else
   {
    if(stringcasecmp(name, "from")==0 ||
       stringcasecmp(name, "sender")==0 ||
       stringcasecmp(name, "reply-to")==0 || 
       stringcasecmp(name, "to")==0 ||
       stringcasecmp(name, "cc")==0 ||
       stringcasecmp(name, "bcc")==0)
      address_list_to_imap(header);
    else
      dbimap4d.send_string(header);
   }
 }

/*******************************************************************\

Function: fetch_set_commandt::fetch_envelope

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::fetch_envelope(rfcmessaget &rfcm)
 {
  dbimap4d.begin_list();
  dbimap4d.list_item(); send_header(rfcm, "date");
  dbimap4d.list_item(); send_header(rfcm, "subject");
  dbimap4d.list_item(); send_header(rfcm, "from");
  dbimap4d.list_item(); send_header(rfcm, "sender");
  dbimap4d.list_item(); send_header(rfcm, "reply-to");
  dbimap4d.list_item(); send_header(rfcm, "to");
  dbimap4d.list_item(); send_header(rfcm, "cc");
  dbimap4d.list_item(); send_header(rfcm, "bcc");
  dbimap4d.list_item(); send_header(rfcm, "in-reply-to");
  dbimap4d.list_item(); send_header(rfcm, "message-id");
  dbimap4d.end_list();
 }

/*******************************************************************\

Function: fetch_set_commandt::fetch_body

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::fetch_body(const std::string &what)
 {
  if(stringcasecmp(what, "ENVELOPE")==0)
   {
    if(!dbimap4d.get_header(nr-1))
     {
      dbimap4d.list_item();
      dbimap4d.out << what;
      dbimap4d.list_item();
      fetch_envelope(dbimap4d.rfcm);
     }
   }
  else if(stringcasecmp(what, "BODY")==0 ||
          stringcasecmp(what, "BODYSTRUCTURE")==0)
    fetch_bodystructure(what);
  else if(strnicmp(what.c_str(), "BODY", 4)==0)
    fetch_body2(what, "");
  else if(stringcasecmp(what, "RFC822")==0)
    fetch_body2("BODY[]", what);
  else if(stringcasecmp(what, "RFC822.TEXT")==0)
    fetch_body2("BODY[TEXT]", what);
  else if(stringcasecmp(what, "RFC822.HEADER")==0)
    fetch_body2("BODY.PEEK[HEADER]", what);
  else if(stringcasecmp(what, "RFC822.SIZE")==0)
   {
    if(!dbimap4d.get_header(nr-1))
     {
      dbimap4d.list_item();
      dbimap4d.out << what;
      dbimap4d.list_item();
      dbimap4d.out << (dbimap4d.rfcm.headerlen+dbimap4d.fi.len);
     }
   }
 }

/*******************************************************************\

Function: fetch_set_commandt::fetch

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void fetch_set_commandt::fetch(const std::string &what)
 {
  if(stringcasecmp(what, "ALL")==0)
   {
    fetch("FLAGS");
    fetch("INTERNALDATE");
    fetch("RFC822.SIZE");
    fetch("ENVELOPE");
   }
  else if(strnicmp(what.c_str(), "BODY", 4)==0 ||
          strnicmp(what.c_str(), "RFC822", 6)==0 ||
          stringcasecmp(what, "ENVELOPE")==0 ||
          stringcasecmp(what, "BODYSTRUCTURE")==0)
   {
    fetch_body(what);
   }
  else if(stringcasecmp(what, "FAST")==0)
   {
    fetch("FLAGS");
    fetch("INTERNALDATE");
    fetch("RFC822.SIZE");
   }
  else if(stringcasecmp(what, "FLAGS")==0)
   {
    dbimap4d.list_item();
    dbimap4d.out << "FLAGS";
    dbimap4d.list_item();
    dbimap4d.begin_list();
    if(dbimap4d.fi.flags.imapflags&DBOXIMAP4_FLAG_Answered)
     { dbimap4d.list_item(); dbimap4d.out << "\\Answered"; }
    if(dbimap4d.fi.flags.imapflags&DBOXIMAP4_FLAG_Flagged)
     { dbimap4d.list_item(); dbimap4d.out << "\\Flagged"; }
    if(dbimap4d.fi.flags.deleted) 
     { dbimap4d.list_item(); dbimap4d.out << "\\Deleted"; }
    if(dbimap4d.fi.flags.seen)
     { dbimap4d.list_item(); dbimap4d.out << "\\Seen"; }
    if(dbimap4d.fi.flags.imapflags&DBOXIMAP4_FLAG_Draft)
     { dbimap4d.list_item(); dbimap4d.out << "\\Draft"; }
    if(FALSE)
     { dbimap4d.list_item(); dbimap4d.out << "\\Recent"; }
    dbimap4d.end_list();
   }
  else if(stringcasecmp(what, "FULL")==0)
   {
    fetch("FLAGS");
    fetch("INTERNALDATE");
    fetch("RFC822.SIZE");
    fetch("ENVELOPE");
    fetch("BODY");
   }
  else if(stringcasecmp(what, "INTERNALDATE")==0)
   {
    dbimap4d.list_item();
    dbimap4d.out << "INTERNALDATE";
    dbimap4d.list_item();
    dbimap4d.print_date(dbimap4d.fi.received);
   }
  else if(stringcasecmp(what, "UID")==0)
   {
    dbimap4d.list_item();
    dbimap4d.out << "UID";
    dbimap4d.list_item();
    dbimap4d.out << dbimap4d.fi.id;
   }
 }

/*******************************************************************\

Function: fetch_set_commandt::msg

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool fetch_set_commandt::msg()
 {
  std::string item;
  bool uid_already_done=FALSE;

  dbimap4d.get_folderindex(nr-1);

  argnr=uid?2:1;

  dbimap4d.out << "* " << nr << " " 
               << (uid?dbimap4d.arguments[0].data:dbimap4d.command)
               << " ";

  dbimap4d.begin_list();

  for(; argnr<dbimap4d.arguments.size(); argnr++)
   {
    if(dbimap4d.arguments[argnr].begin_list ||
       dbimap4d.arguments[argnr].end_list) 
     {
     }
    else
     {
      item=dbimap4d.arguments[argnr].data;
      upstring(item);
      fetch(item);
      if(stringcasecmp(item, "UID")==0) uid_already_done=TRUE;
     }
   }  

  if(uid && !uid_already_done)
    fetch("UID");   

  dbimap4d.end_list();

  dbimap4d.out << "\r\n";

  return FALSE;
 }

