/*

  DBOX Module
  Copyright (C) 1994-2003 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <fstream>

#include <stringtools.h>
#include <dblib/os_dep_file.h>
#include <dbclient/lock.h>

#include "dbimap4d.h"
#include "mod-utf7.h"

/*******************************************************************\

Function: dbimap4dt::dosubscribe

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::dosubscribe(bool subscribe)
 {
  if(arguments.size()!=1)
   {
    error("wrong number of arguments");
    return;
   }  

  std::string mailbox_name=decode_mod_utf7(arguments[0].data);

   {
    lockt folder_lock("folders", user, "dbimap4t::dosubscribe");

    get_folder_list();

    unsigned i;
    for(i=0; i<folders.size(); i++)
      if(mailbox_name==folders[i].name)
        break;

    if(i==folders.size())
     {
      error("no such mailbox");
      return;
     }

    if(folders[i].subscribed!=subscribe)
     {
      folders[i].subscribed=subscribe;

      std::string outfilename=
        build_path(get_user_path(), "subscribed");

      std::ofstream out(outfilename.c_str(),
                        std::ios::out|std::ios::trunc);

      for(i=0; i<folders.size(); i++)
        if(folders[i].subscribed)
          out << 'F' << i << "\n";
     }
   }
 }

/*******************************************************************\

Function: dbimap4dt::get_folder_list

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::get_folder_list()
 {
  std::string filename=build_path(get_user_path(), "folders");

  folders.clear();

  std::ifstream in(filename.c_str());

  if(!in)
   {
    folder_et folder_e;
    folder_e.name="INBOX";
    folder_e.subscribed=true;
    folders.push_back(folder_e);
   }
  else
   {
    char line[TEXTLEN], *tptr, ch;

    while(in.get(line, TEXTLEN-1))
     {
      in.get(ch); // read \n
      tptr=strchr(line, '\n'); if(tptr!=NULL) *tptr=0;
      tptr=strchr(line, '\r'); if(tptr!=NULL) *tptr=0;

      folder_et folder_e;
      folder_e.name=line;
      folders.push_back(folder_e);
     }
   }

  get_subscribed();
 }

/*******************************************************************\

Function: dbimap4dt::get_subscribed

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::get_subscribed()
 {
  std::string filename=build_path(get_user_path(), "subscribed");

  std::ifstream in(filename.c_str());

  if(!in) return;

  char line[TEXTLEN], *tptr, ch;

  while(in.get(line, TEXTLEN-1))
   {
    in.get(ch); // read \n
    tptr=strchr(line, '\n'); if(tptr!=NULL) *tptr=0;
    tptr=strchr(line, '\r'); if(tptr!=NULL) *tptr=0;

    if(line[0]=='F')
     {
      unsigned i=atol(line+1);
      if(i<folders.size())
        folders[i].subscribed=TRUE;
     }
   }
 }

/*******************************************************************\

Function: dbimap4dt::matchfolder

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbimap4dt::matchfolder(const std::string &mask,
                            const std::string &folder)
 {
  return matchfolder(mask.c_str(), folder.c_str());
 }

/*******************************************************************\

Function: dbimap4dt::matchfolder

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbimap4dt::matchfolder(const char *mask, const char *folder)
 {
  while(*mask!=0)
   {
    switch(*mask)
     {
     case '*':
      mask++;
      if(*mask==0) return TRUE;
      while(*folder!=0)
       {
        if(matchfolder(mask, folder)) return TRUE;
        folder++;
       }
      return FALSE;

     case '%':
      mask++;

      if(*mask==0)
       {
        for(; *folder!=0; folder++)
          if(*folder=='/') return FALSE;
        return TRUE;
       }

      while(*folder!=0)
       {
        if(matchfolder(mask, folder)) return TRUE;
        if(*folder=='/') return FALSE;
        folder++;
       }
      return FALSE;

     default:
      if(*mask!=*folder)
        return FALSE;

      folder++;
     }

    mask++;
   }

  return (*folder==0);
 }

/*******************************************************************\

Function: dbimap4dt::dolsublist

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::dolsublist()
 {
  bool lsub=(stringcasecmp(command, "LSUB")==0);

  if(arguments.size()!=2)
   {
    error("please give arguments");
    return;
   }

  if(arguments[1].data=="")
   {
    out << "* " << command << " (\\Noselect) \"/\" \"\"\r\n";
   }
  else
   {
    get_folder_list();

    for(unsigned i=0; i<folders.size(); i++)
      if((folders[i].name!="") &&
         (folders[i].subscribed || !lsub))
       {
        std::string foldername=encode_mod_utf7(folders[i].name);

        if(matchfolder(arguments[1].data, foldername))
         {
          //folder_grantst grants;
          //std::string subfolder=i2string(i);

          //::get_folder_grants(up.name, subfolder, up, gp, grants);
          //if(grants.lookup)
          // {
            out << "* " << command << " () \"/\" ";
            send_string(foldername);
            out << "\r\n";
          // }
         }
       }
   }

  command_ok=true;
 }

