/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <sstream>
#include <iostream>

#include <stringtools.h>
#include <dblib/stream_copy.h>
#include <dblib/rfc1123date.h>
#include <dblib/os_dep_file.h>
#include <dbclient/logstr.h>
#include <dbclient/client.h>
#include <dbsmtpd/folder_path.h>

#include "dbimap4d.h"
#include "mod-utf7.h"

/*******************************************************************\

Function: dbimap4dt::get_folder_grants

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::get_folder_grants(const mailfoldert &src_mailfolder,
                                  folder_grantst &dest)
 {
  // TODO
  dest.del=dest.lookup=dest.read=dest.write=dest.insert=
  dest.create=dest.administer=dest.keepseen=true;
 }

/*******************************************************************\

Function: dbimap4dt::channel_lost

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::channel_lost()
 {
  logstr << "channel lost";
  logstr.log("dbimap4dt::channel_lost");
 }

/*******************************************************************\

Function: dbimap4dt::find_folder

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

dbimap4dt::folderst::iterator dbimap4dt::find_folder
 (const std::string &name)
 {
  for(folderst::iterator it=folders.begin();
      it!=folders.end();
      it++)
    if(it->name==name)
      return it;

  return folders.end();
 }

/*******************************************************************\

Function: dbimap4dt::get_user_path

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

std::string dbimap4dt::get_user_path()
 {
  return build_path(client.config["maildir"], user);
 }

/*******************************************************************\

Function: dbimap4dt::check_mailbox_name

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbimap4dt::check_mailbox_name(const std::string &mailbox_name)
 {
  for(unsigned i=0; i<mailbox_name.size(); i++)
   {
    register char ch=mailbox_name[i];

    if(isascii(ch) && iscntrl(ch))
      return TRUE;
   }

  return FALSE;
 }

/*******************************************************************\

Function: dbimap4dt::send_string

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::send_string(const char *string)
 {
  const char *tptr;

  if(string[0]==0)
   {
    std::cout << "NIL";
    return;
   }

  for(tptr=string; *tptr!=0; tptr++)
    if(*tptr=='\n' || *tptr=='\r' || !isascii(*tptr))
     {
      std::cout << '{' << strlen(string) << "}\r\n" << string;
      return;
     }

  std::cout << '"';

  for(tptr=string; *tptr!=0; tptr++)
   {
    if(*tptr=='\\' || *tptr=='"') std::cout << '\\';
    std::cout << char(*tptr);
   }

  std::cout << '"';
 }

/*******************************************************************\

Function: dbimap4dt::send_string

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::send_string(const std::string &string)
 {
  send_string(string.c_str());
 }

/*******************************************************************\

Function: dbimap4dt::process_mailbox_name

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbimap4dt::process_mailbox_name(const std::string &src_mailbox_name,
                                     mailfoldert &dest)
 {
  std::string new_mailbox_name=decode_mod_utf7(src_mailbox_name);

  if(stringcasecmp(new_mailbox_name, "INBOX")==0)
    new_mailbox_name="INBOX";

  std::string maildir=client.config["maildir"];
  std::string name;

  folder_name(maildir, user, new_mailbox_name, name);

  if(name.empty())
   {
    out << tag << " NO no such mailbox\r\n";
    command_ok=FALSE;
    return true;
   }

  dest.set_user(user);
  dest.set_folder_name(name);

  return false;
 }

/*******************************************************************\

Function: dbimap4dt::parse_date

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

#define PARSE_DATE_SKIP(p,c) while(*(p)!=(c) && *(p)!=0) (p)++; \
                             if(*(p)==0) return; else (p)++

void dbimap4dt::parse_date(const std::string &date, struct tm &tm)
 {
  const char *ptr=date.c_str();

  if(ptr[0]==0) return;
  tm.tm_mday=atoi(ptr);
  PARSE_DATE_SKIP(ptr, '-');

  unsigned mon=0;
  for(mon=0; mon<12; mon++)
    if(strnicmp(rfc_months[mon], ptr, 3)==0)
      break;

  if(mon==12) return;
  tm.tm_mon=mon;

  PARSE_DATE_SKIP(ptr, '-');

  tm.tm_year=atoi(ptr)-1900;
  PARSE_DATE_SKIP(ptr, ' ');

  tm.tm_hour=atoi(ptr);
  PARSE_DATE_SKIP(ptr, ':');
    
  tm.tm_min=atoi(ptr);
  PARSE_DATE_SKIP(ptr, ':');
    
  tm.tm_sec=atoi(ptr);
 }

/*******************************************************************\

Function: dbimap4dt::parse_date

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

time_t dbimap4dt::parse_date(const std::string &date)
 {
  struct tm tm;

  time_t t;
  time(&t);

  tm=*localtime(&t);

  parse_date(date, tm);
  
  return mktime(&tm);
 }

/*******************************************************************\

Function: dbimap4dt::print_date

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::print_date(const time_t &t)
 {
  struct tm *tm;

  tm=localtime(&t);

  signed long int tzdiff;

  #ifdef __CYCWIN32__
  tzdiff=0;
  #else
  tzdiff=-tm->tm_gmtoff;
  //if(tm->tm_isdst>0) tzdiff+=60*60;
  #endif

  printf("\"%2d-%s-%04d %02d:%02d:%02d %c%02ld%02ld\"",
    tm->tm_mday, rfc_months[tm->tm_mon], tm->tm_year+1900,
    tm->tm_hour, tm->tm_min, tm->tm_sec, 
    tzdiff<=0?'+':'-',
    (labs(tzdiff)/60/60)%100, (labs(tzdiff)/60)%60);
 }

/*******************************************************************\

Function: dbimap4dt::begin_list

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::begin_list()
 {
  liststack.push_back(TRUE);
  std::cout << '(';
 }

/*******************************************************************\

Function: dbimap4dt::end_list

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::end_list()
 {
  if(liststack.size()==0) abort();
  std::cout << ')';
  liststack.pop_back();
 }

/*******************************************************************\

Function: dbimap4dt::list_item

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::list_item()
 {
  if(liststack.size()==0) abort();

  if(liststack.back())
    liststack.back()=FALSE;
  else
    std::cout << ' ';
 }

/*******************************************************************\

Function: dbimap4dt::dbimap4dt

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

dbimap4dt::dbimap4dt(std::istream &_in, std::ostream &_out):
  popt(_in, _out)
 {
  command_ok=logout=readonly=false; 
  protocol="imap4";
  mailbox_selected=false;
  messages=0;
  uidvalidity=1;
 }

/*******************************************************************\

Function: dbimap4dt::showstatus

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::showstatus()
 {
  folderindext fi;
  dword exists=0, recent=0;
  largest_id=0;
  messages=0;

  if(mailbox_selected && index>=0 && header>=0 && !logout)
   {
    lseek(index, 0, SEEK_SET);

    while(read(index, &fi, sizeof fi)==sizeof fi)
     {
      exists++;
      messages++;
      if(fi.id>largest_id) largest_id=fi.id;
     }

    std::cout << "* " << exists << " EXISTS\r\n";
    std::cout << "* " << recent << " RECENT\r\n";
   }
 }

/*******************************************************************\

Function: dbimap4dt::status_update

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::status_update()
 {
  time_t t;

  if(!authenticated || !mailbox_selected) return;

  time(&t);

  #ifdef DBOX_STANDALONE
  if(mailbox=="INBOX")
    movemail();
  #endif

  // if(t>(last_status_update+5))
   {
    struct stat statbuf;
    std::string filename=build_path(mailfolder.get_path(), "index");

    if(stat(filename.c_str(), &statbuf)==0 && 
       statbuf.st_size!=last_index_file_size)
     {
      re_open();
      showstatus();
      last_index_file_size=statbuf.st_size;
      last_status_update=t;
     }
   }

  if(t>(last_status_update+60))
   {
    showstatus();
    last_status_update=t;
   }
 }

/*******************************************************************\

Function: dbimap4dt::parse_args

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::parse_args()
 {
  const char *tptr=argument_string.c_str();

  while(*tptr!=0)
   {
    argumentt argument;

    switch(*tptr)
     {
     case '(':
      tptr++;
      argument.begin_list=TRUE;
      break;

     case ')':
      tptr++;
      argument.end_list=TRUE;
      break;

     case '"': // quoted
      tptr++;
       {
        bool quoted=FALSE;
        while(tptr[0]!=0 && (tptr[0]!='"' && !quoted))
         {
          if(!quoted && tptr[0]=='\\')
            quoted=TRUE;
          else
           {
            quoted=FALSE;
            argument.data+=tptr[0];
           }
          tptr++;
         }
       }

      if(tptr[0]=='"') tptr++;
      break;

     case '{': // literal
      tptr++;

      argument.literal=TRUE;
      argument.literal_size=atol(tptr);

      while(tptr[0]!=0 && tptr[0]!='}')
        tptr++;

      if(tptr[0]=='}') tptr++;
      break;

     default: // "normal"
      while(tptr[0]!=0 && tptr[0]!=' ' && tptr[0]!=')')
        argument.data+=*(tptr++);
     }

    if(*tptr==' ') tptr++;
    arguments.push_back(argument);
   }
 }

/*******************************************************************\

Function: dbimap4dt::parse

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::parse()
 {
  const char *tptr=commandline.c_str();

  tag="";
  while(*tptr!=0 && *tptr!=' ')
    tag+=*(tptr++);

  if(*tptr==' ') tptr++;

  command="";
  while(*tptr!=0 && *tptr!=' ')
    command+=*(tptr++);

  if(*tptr==' ') tptr++;

  argument_string=tptr;

  upstring(command);

  arguments.clear();
  parse_args();
 }

/*******************************************************************\

Function: dbimap4dt::get_and_parse_command

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbimap4dt::get_and_parse_command()
 {
  unsigned args;

  if(getcommand()) return TRUE;

  #ifdef DEBUG
  #ifndef DBOX_STANDALONE
  log_serv.printf("debug", "%d >%s<", getpid(), commandline.c_str());
  #endif
  #endif

  parse();

  args=arguments.size();

  if(args==0) return FALSE;
  
  while(arguments[args-1].literal)
   {
    std::cout << "+ send literal\r\n";

     {
      std::ostringstream data;
      if(filecopy(std::cin, data, arguments[args-1].literal_size))
        return TRUE;

      arguments[args-1].data=data.str();
      assert(arguments[args-1].data.size()==arguments[args-1].literal_size);
     }    

    if(getcommand()) return TRUE;

    #ifdef DEBUG
    #ifndef DBOX_STANDALONE
    log_serv.printf("debug", ">%s<", commandline.c_str());
    #endif
    #endif

    if(commandline=="") return FALSE;
    if(commandline[0]!=' ') return FALSE;

    argument_string=(commandline.c_str()+1); // skip blank

    if(argument_string!="")   
      parse_args();

    if(args==arguments.size()) // nothing new
      return FALSE;

    args=arguments.size();
   }

  return FALSE;
 }
