/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: rename a folder

*/

#include <fstream>

#include <stringtools.h>
#include <dboxpath.h>
#include <lock.h>
#include <dblib/os_dep_file.h>

#include "mod-utf7.h"
#include "dbimap4d.h"

/*******************************************************************\

Function: dbimap4dt::dorename

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbimap4dt::dorename()
 {
  if(arguments.size()!=2)
   {
    error("give two mailbox names");
    return;
   }

  std::string old_mailbox_name=decode_mod_utf7(arguments[0].data),
              new_mailbox_name=decode_mod_utf7(arguments[1].data);

  if(old_mailbox_name=="" ||
     new_mailbox_name=="")
   {
    error("give a non-empty mailbox names");
    return;
   }
  
  if(stringcasecmp(old_mailbox_name, "INBOX")==0)
   {
    error("I won't rename INBOX");
    return;
   }
  
  if(stringcasecmp(new_mailbox_name, "INBOX")==0)
   {
    error("folder already exists");
    return;
   }
  
  if(check_mailbox_name(new_mailbox_name))
   {
    error("new mailbox name is not valid");
    return;
   }

   {
    lockt folders_lock("folders", user, "dbimap4t::dorename");

    get_folder_list();

    if(find_folder(new_mailbox_name)!=folders.end())
     {
      error("mailbox already exists");
      return;
     }

     {
      folderst::iterator it=find_folder(old_mailbox_name);

      if(it==folders.end())
       {
        error("no such mailbox");
        return;
       }

      it->name=new_mailbox_name;
     }

     {
      std::string outfilename=build_path(get_user_path(), "folders");
      std::ofstream outfile(outfilename.c_str(), std::ios::out);

      for(folderst::const_iterator it=folders.begin();
          it!=folders.end(); it++)
        outfile << it->name << "\n";
     }
   }
 }
