/*******************************************************************\

Module:

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#include <vector>

#include "os_dep_socket.h"
#include "os_dep_file.h"
#include "socketstr.h"

#ifndef	WIN32
# ifndef  O_NONBLOCK
#  define O_NONBLOCK	O_NDELAY
# endif
# ifdef	IPPROTO_IP
#  ifndef  SOL_IP
#   define SOL_IP	IPPROTO_IP
#  endif // !SOL_IP
# endif	 // IPPROTO_IP
#endif	 // !WIN32

#if 0
socketstrt::TCPStream(TCPSocket &server, int size, bool throwflag, timeout_t to) :
	Socket(accept(server.so, NULL, NULL))
	,streambuf()
#ifdef	HAVE_OLD_IOSTREAM
	,iostream()
#else
	,iostream((streambuf *)this)
#endif
	,bufsize(0)
	,gbuf(NULL)
	,pbuf(NULL)
{
	tpport_t port;

#ifdef	HAVE_OLD_IOSTREAM
	init((streambuf *)this);
#endif

	timeout = to;
	setError(throwflag);
	InetHostAddress host = getPeer(&port);
	if(!server.onAccept(host, port))
	{
		endSocket();
		error(errConnectRejected);
		clear(ios::failbit | rdstate());
		return;
	}

	allocate(size);
	Socket::state = CONNECTED;
}

socketstrt::TCPStream(const InetHostAddress &host, tpport_t port, int size, bool throwflag, timeout_t to) :
	Socket(AF_INET, SOCK_STREAM, 0), 
	streambuf(), 
#ifdef	HAVE_OLD_IOSTREAM
	iostream(),
#else
        iostream((streambuf *)this),
#endif
	bufsize(0),gbuf(NULL),pbuf(NULL)
{
#ifdef	HAVE_OLD_IOSTREAM
	init((streambuf *)this);
#endif
	timeout = to;
	setError(throwflag);
	connect(host, port, size);
}

void socketstrt::connect(const InetHostAddress &host, tpport_t port, int size)
{
}
#endif

void socketstrt::endStream(void)
 {
  if(bufsize) sync();
  delete [] gbuf;
  delete [] pbuf;

  if(so>=0)
    ::close(so);

  init();

  clear();
 }

void socketstrt::allocate(int size)
 {
  if(size<2)
   {
    bufsize = 1;
    gbuf = pbuf = 0;
    return;
   }

  gbuf = new char[size];
  pbuf = new char[size];

  if(!pbuf || !gbuf)
    throw excpt("out of memory");

  bufsize = size;
  clear();

  #if (defined(__GNUC__) && (__GNUC__ < 3)) && !defined(WIN32) && !defined(STLPORT)
  setb(gbuf, gbuf + size, 0);
  #endif

  setg(gbuf, gbuf + size, gbuf + size);
  setp(pbuf, pbuf + size);
 }

int socketstrt::doallocate()
 {
  if(bufsize) return 0;

  allocate(1);
  return 1;
 }

int socketstrt::uflow()
 {
  int ret = underflow();

  if (ret == EOF)
    return EOF; 

  if (bufsize != 1)
    gbump(1); 

  return ret; 
 }

int socketstrt::underflow()
 {
  int rlen = 1;
  unsigned char ch;

  if(bufsize == 1)
   {
    rlen = ::read(so, (char *)&ch, 1);

    if(rlen<1)
     {
      if(rlen<0)
       {
	clear(std::ios::failbit | rdstate());
	throw excpt("socket input failed");
       }

      return EOF;
     }

    return ch;
   }

  if(!gptr())
    return EOF;

  if(gptr()<egptr())
    return (unsigned char)*gptr();

  rlen=(gbuf+bufsize)-eback();

  rlen=::read(so, (char *)eback(), rlen);

  if(rlen<1)
   {
    if(rlen<0)
     {
      clear(std::ios::failbit | rdstate());
      throw excpt("socket input failed");
     }

    return EOF;
   }

  setg(eback(), eback(), eback() + rlen);

  return (unsigned char) *gptr();
 }

int socketstrt::sync()
 {
  overflow(EOF);
  setg(gbuf, gbuf + bufsize, gbuf + bufsize);
  return 0;
 }

int socketstrt::overflow(int c)
 {
  unsigned char ch;
  int rlen, req;

  if(bufsize == 1)
   {
    if(c == EOF)
      return 0;

    ch = (unsigned char)(c);

    rlen = ::write(so, (const char *)&ch, 1);

    if(rlen < 1)
     {
      if(rlen < 0)
       {
	clear(std::ios::failbit | rdstate());
        throw excpt("socket output failed");
       }
      return EOF;
     }
    else
      return c;
   }		

  if(!pbase())
    return EOF;

  req = pptr() - pbase();

  if(req)
   {
    rlen = ::write(so, (const char *)pbase(), req);

    if(rlen < 1)
     {
      if(rlen < 0)
       {
	clear(std::ios::failbit | rdstate());
        throw excpt("socket output failed");
       }

      return EOF;
     }

    req -= rlen;
   }

  // if write "partial", rebuffer remainder
	
  if(req)
    memmove(pbuf, pptr() + rlen, req);

  setp(pbuf, pbuf + bufsize);
  pbump(req);

  if(c != EOF)
   {
    *pptr() = (unsigned char)c;
    pbump(1);
   }

  return c;
 }

void socketstrt::open(const std::string &host, unsigned port, int buffer)
 {
  endStream();

  // look up the name

  std::vector<struct sockaddr_in> snlist;

   {
    struct sockaddr_in sn;

    memset(&sn, 0, sizeof(sn));
    sn.sin_family=AF_INET;
    sn.sin_port=htons(port);

    struct hostent *hp=gethostbyname(host.c_str());

    if(!hp) // failed
     {
      struct in_addr defaddr;

      if(!inet_aton(host.c_str(), &defaddr)) // see if it is 1.2.3.4
       {
        memcpy(&sn.sin_addr, &defaddr, hp->h_length);
        snlist.push_back(sn);
       }
     }
    else
     {
      if(hp->h_length>(int)sizeof(sn.sin_addr))
        hp->h_length = sizeof(sn.sin_addr);

      for(unsigned i=0; hp->h_addr_list[i]!=NULL; i++)
       {
        memcpy(&sn.sin_addr, hp->h_addr_list[i], hp->h_length);
        snlist.push_back(sn);
       }
     }
   }

  if(snlist.size()==0)
    throw excpt("unknown host: "+host);

  if((so=::socket(AF_INET, SOCK_STREAM, 0))<0)
    throw excpt("failed to create socket");

  bool connected=false;

  for(unsigned i=0; i<snlist.size(); i++)
   {
    int rtn=::connect(so, (struct sockaddr *)&snlist[i], sizeof(struct sockaddr_in));

    if(!rtn)
     {
      connected=true;
      break;
     }

    if(errno==EINPROGRESS)
     {
      fd_set fds;
      struct timeval to;

      FD_ZERO(&fds);
      FD_SET(so, &fds);
      to.tv_sec = timeout / 1000;
      to.tv_usec = timeout % 1000 * 1000;
			
      // timeout check for connect completion

      if(::select(so+1, NULL, &fds, NULL, &to)<1)
	continue;

      long sockopt;
      socklen_t len = sizeof(sockopt);

      getsockopt(so, SOL_SOCKET, SO_ERROR, (char *)&sockopt, &len);
      if(!sockopt)
       {
	connected=true;
	break;
       }

      ::close(so);

      if((so=::socket(AF_INET, SOCK_STREAM, 0))<0)
        throw excpt("failed to create socket");
     }
   }

  if(!connected)
   {
    ::close(so);
    throw excpt("failed to connect to "+host);
   }

  allocate(buffer);
 }
	
void socketstrt::close(void)
 {
  endStream();
 }
