/*******************************************************************\

Module:

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#ifndef	DBLIB_SOCKET_STR_H
#define	DBLIB_SOCKET_STR_H

#include <iostream>

#include "os_dep_socket.h"
#include "excpt.h"

EXCPT(socket_excpt);

class socketstrt: protected std::streambuf, public std::iostream
 {
 private:
  int doallocate();

  friend socketstrt& crlf(socketstrt&);
  friend socketstrt& lfcr(socketstrt&);

 protected:
  unsigned long timeout;
  int bufsize;
  char *gbuf, *pbuf;

  void allocate(int size);
  void endStream();
  int underflow();
  int uflow();
  int overflow(int ch);
  void connect(const std::string &host, unsigned port, int size);

  std::iostream *tcp()
   { return ((std::iostream *)this); }
   
  int so; // socket descriptor
  
  void init()
   {
    timeout=0;
    so=-1;
    gbuf=pbuf=NULL;
    bufsize=0;
   }

 public:
  socketstrt():std::iostream((std::streambuf *)this)
   {
    init();
   }

  socketstrt(const std::string &host, unsigned port, int buffer = 512):
    std::iostream((std::streambuf *)this)
   {
    init();
    open(host, port, buffer);
   }

  inline void setTimeout(unsigned long _timeout)
   { timeout = _timeout; }

  virtual ~socketstrt()
   { endStream(); }

  int sync();

  int getBufferSize() const
   { return bufsize; }

  void open(const std::string &host, unsigned port, int buffer = 512);

  void close();
 };

#endif
