/*******************************************************************\

Module:

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#include "os_dep_setjmp.h"
#include "os_dep_process.h"

#include "str_getline.h"

/*******************************************************************\

Function: str_getline_main

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

std::istream &str_getline_main(std::istream &in, std::string &dest,
                               bool crlf)
 {
  #define BUFSIZE 4096

  char buffer[BUFSIZE]="";
  char ch;
  unsigned i=0;

  dest.erase();

  while(in)
   {
    in.get(ch);

    if(ch=='\n')
      break;

    if(crlf && ch=='\r')
      continue; // ignore it

    buffer[i++]=ch;
    buffer[i]=0;

    if(i==(BUFSIZE-1))
     {
      dest+=buffer;
      i=0;
      buffer[i]=0;
     }
   }

  dest+=buffer;

  return in;
 }

/*******************************************************************\

Function: str_getline

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

std::istream &str_getline(std::istream &in, std::string &dest)
 {
  return str_getline_main(in, dest, false);
 }

/*******************************************************************\

Function: str_getline_crlf

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

std::istream &str_getline_crlf(std::istream &in, std::string &dest)
 {
  return str_getline_main(in, dest, true);
 }

/*******************************************************************\

Function: str_getline_alarm

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

jmp_buf str_getline_jmp;

void str_getline_alarm(int signal)
 {
  longjmp(str_getline_jmp, 1);
 }

/*******************************************************************\

Function: str_getline

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

std::istream &str_getline(std::istream &in, std::string &dest,
                          bool crlf,
                          unsigned timeout, bool &got_timeout)
 {
  got_timeout=false;

  if(setjmp(str_getline_jmp)!=0)
   {
    got_timeout=true;
    return in;
   }

  signal(SIGALRM, str_getline_alarm);

  alarm(timeout);
  std::istream &result=str_getline_main(in, dest, crlf);
  alarm(0);

  return result;
 }

