/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: base class for remote mailbox access

*/

#include <dblib/os_dep_file.h>

#include <tools.h>
#include <log_serv.h>
#include <logfile.h>
#include <flags.h>
#include <dboxpath.h>
#include <dblib/getpeer.h>
#include "pop.h"

/*******************************************************************\

Function:

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

dboxpopt::dboxpopt()
 {
  transfer=FALSE;
  protocol="unknown";
  timeout_minutes=10;
 }

/*******************************************************************\

Function:

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

#define MAX_LINE_LENGTH 10000

bool dboxpopt::getcommand()
 {
  char *tptr;
  char line[MAX_LINE_LENGTH+1];

  #ifndef WIN32
  alarm(timeout_minutes*60);
  #endif

  fflush(stdout);
  if(fgets(line, MAX_LINE_LENGTH, stdin)==NULL) return TRUE; 

  #ifndef WIN32
  alarm(0);
  #endif

  tptr=strchr(line, '\r'); if(tptr!=NULL) *tptr=0;
  tptr=strchr(line, '\n'); if(tptr!=NULL) *tptr=0;

  commandline=line;
 
  return FALSE;
 }

/*******************************************************************\

Function:

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dboxpopt::login(const std::string &password)
 {
  if(up.name[0]==0 || strchr(up.name, '/')!=NULL)
    return TRUE;

  if(up.read())
   {
    log_serv.printf("login", "Invalid user name %s from %s",
                    up.name, peer.c_str());
    return TRUE;
   }

  snprintf(userlogfile, DIRLEN-1, "%slogfile", up.getpath());

  if(verifylogin(&up, password.c_str()))
   {
    logfileprintf(userlogfile, "login", 
                  "wrong %s password from %s",
                  protocol.c_str(), peer.c_str());
    log_serv.printf("login", "%s: wrong password from %s",
                    up.name, peer.c_str());
    return TRUE;
   }

  return login2();
 }

/*******************************************************************\

Function:

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dboxpopt::login2()
 {
  userpt up2;
  get_profiles(up, up2, gp);

  if(!get_internalflag(up, up2, gp, LOGIN_POP3))
   {
    logfileprintf(userlogfile, "login", "%s access denied", protocol.c_str());
    log_serv.printf("login", "%s: access denied", up.name);
    return TRUE;
   }

  logfileprintf(userlogfile, "login", "Successfull %s login from %s",
                protocol.c_str(), peer.c_str());

  log_serv.printf("login", "%s: Successfull %s login from %s", 
    up.name, protocol.c_str(), peer.c_str());

  transfer=TRUE;

  return FALSE;
 }

/*******************************************************************\

Function:

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dboxpopt::getpeer()
 {
  peer=::getpeer(0);
 }

/*******************************************************************\

Function:

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dboxpopt::docrunch(const char *username)
 {
  char filename[DIRLEN];
  FILE *outfile;

  snprintf(filename, DIRLEN, BOXTODODIR "crunch.%s", config.boxdir, username);
  outfile=fopen(filename, "wt");
  if(outfile!=NULL) fclose(outfile);
 }

/*******************************************************************\

Function:

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

#ifdef DBOX_STANDALONE
void dboxpopt::movemail()
 {
  ::movemail(up);
 }
#endif
