/*

  DBOX Module
  Copyright (C) 1994-2003 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <string>
#include <iostream>

#include <dblib/os_dep_file.h>
#include <dbclient/client.h>
//#include <tools.h>
//#include <folder.h>
#include <stringtools.h>
#include <dbclient/logstr.h>
//#include <logfile.h>
#include <dboxversion.h>
//#include <messageid.h>

#ifndef DBOX_STANDALONE
#include <md5.h>
#endif

#include "pop.h"
#include "dbpop3d.h"

class dbpop3dt:public popt
 {
 public:
  std::string command, args;
  std::vector<dword> messages;
  std::vector<bool> deleted;
  dword lastmsg;
  std::string timestamp;

  void dolist(dword arg, bool only_parameter, bool uidl);
  bool logged_in(); 
  void dbpop3d();
  bool do_command();
  bool do_command_not_authenticated();
  bool do_command_authenticated();
  void parse();
  void retr_top(dword nr, const char *toparg);

  virtual void timeout();

  dbpop3dt(std::istream &_in, std::ostream &_out);
 };

/*******************************************************************\

Function: dbpop3t::timeout

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbpop3dt::timeout()
 {
  out << "-ERR logout, timeout\r\n";
 }

/*******************************************************************\

Function: dbpop3dt::dbpop3dt

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

dbpop3dt::dbpop3dt(std::istream &_in, std::ostream &_out):
          popt(_in, _out)
 {
  protocol="pop3";
  lastmsg=0;
 }

/*******************************************************************\

Function: dbpop3dt::parse

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbpop3dt::parse()
 {
  const char *tptr=commandline.c_str();

  while(*tptr==' ' || *tptr=='\t') tptr++;

  command="";
  while(*tptr!=0 && *tptr!=' ' && *tptr!='\t')
    command+=toupper(*(tptr++));

  if(*tptr==' ' || *tptr=='\t')
   {
    while(*tptr==' ' || *tptr=='\t') tptr++;
    args=tptr;
   }
  else
    args="";
 }

/*******************************************************************\

Function: dbpop3dt::retr_top

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbpop3dt::retr_top(dword nr, const char *toparg)
 {
  dword lines;

  //
  // get header and, if necessary, body
  //

  if(toparg==NULL)
    lines=0xffffffff;
  else
    lines=atol(toparg);

  if(toparg==NULL)
    out << "+OK message follows\r\n";
  else
    out << "+OK top of message follows\r\n";

  if(lines!=0)
    get_body(nr);
  else
    get_header(nr);

  // output header

  out.write(rfcm.data, rfcm.headerlen);

  // output body

  if(lines!=0)
   {
    const char *ptr=rfcm.data+rfcm.headerlen;
    char data=0, lastchar=0;
    byte escapestatus=1; // 0=nix, 1=Zeilenanfang, 2=Zeilenanfang+'.'

    while(lines!=0 && ptr<rfcm.endofmail)
     {
      lastchar=data;

      data=*(ptr++);

      switch(escapestatus)
       {
       case 0:
        if(data=='\r' || data=='\n')
          escapestatus=1;
        break;

       case 1:
        if(data=='.')
         {
          escapestatus=2;
          out << '.'; // Hm. Netscape macht aus ... => ..
         }
        else if(data=='\r' || data=='\n')
          escapestatus=1;
        else
          escapestatus=0;
        break;

       case 2:
        if(data=='\r' || data=='\n')
          escapestatus=1;
        else
          escapestatus=0;
        break;
       }

      if(data=='\n' && lastchar!='\r')
        out << '\r';

      out << data;

      if(data=='\n') lines--;
     }
  
    if(data!='\n')
      out << "\r\n";
   }  

  out << ".\r\n";
 }

/*******************************************************************\

Function: dbpop3dt::dolist

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbpop3dt::dolist(dword arg, bool only_parameter, bool uidl)
 {
  if(only_parameter)
   {
    if(deleted[arg-1])
     {
      out << "-ERR message is deleted\r\n";
      return;
     }

    out << "+OK ";
   }
  else
   {
    if(deleted[arg-1])
      return;
   }

  dword nr=messages[arg-1];
  get_folderindex(nr);

  out << arg << " ";

  if(uidl)
   {
    out << fi.id << "."
        << str2crc(fi.mid) << "."
        << fi.received;
   }
  else
   {
    if(!fi.flags.binary)
      out << (fi.len+fi.headerlen);
    else
      out << (fi.headerlen+fi.len*1377/1000);
   }

  out << "\r\n";
 }

/*******************************************************************\

Function: dbpop3dt::logged_in

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbpop3dt::logged_in()
 {
  if(open_mail("0"))
    throw excpt("failed to open mail folder");

  if(index>=0 && header>=0)
   {
    dword idx=0;

    while(read(index, &fi, sizeof fi)==sizeof fi)
     {
      if(fi.check_version())
       {
        logstr << "mailbox " << user << " corrupted";
        logstr.log("dbpop3dt::logged_in");
        throw excpt("sorry, mailbox is corrupted");
       }

      if(!fi.flags.deleted)
        messages.push_back(idx);

      idx++;
     }

    deleted.resize(messages.size(), false);
   }

  return false;
 }

/*******************************************************************\

Function: dbpop3dt::dbpopd3 

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbpop3dt::dbpop3d()
 {
  try
   {
    if(client.config["maildir"]=="")
      throw excpt("Mail system configuration error");

    out << "+OK DBPOPD3 Server " DBOXVERSIONNR " ready\r\n";

    getpeer();

    while(true)
     {
      if(getcommand()) return;
      parse();
      if(do_command()) return;
     }
   }

  catch(excpt e)
   {
    out << "-ERR " << e.what() << "\r\n";
   }
 }

/*******************************************************************\

Function: dbpop3dt::do_command

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbpop3dt::do_command()
 {
  if(authenticated)
    return do_command_authenticated();

  return do_command_not_authenticated();
 }

/*******************************************************************\

Function: dbpop3dt::do_command_not_authenticated

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbpop3dt::do_command_not_authenticated()
 {
  if(command=="QUIT")
   {
    out << "+OK " << client.config["hostname"]
        << " DBOX POP Server (Version " DBOXVERSIONNR ") shutdown.\r\n";

    return true;
   }
  else if(command=="USER")
   {
    user=args;

    if(user=="")
      out << "-ERR invalid user name\r\n";
    else
     {
      dnstring(user);
      out << "+OK please send PASS command\r\n";
     }
   }
  else if(command=="PASS")
   {
    if(user=="")
     {
      out << "-ERR please provide an user name first\r\n";
     }
    else
     {
      if(authenticate(args))
       {
        out << "-ERR invalid password, bye\r\n";
        return true;
       }

      if(logged_in())
        return true;

      out << "+OK Welcome\r\n";
     }
   }
  else if(command=="NOOP")
   {
    if(args!="")
      out << "-ERR Uh?\r\n";
    else
      out << "+OK Gaehn.\r\n";
   }
  else if(command=="APOP")
   {
    out << "-ERR sorry, no APOP supported\r\n";
   }
  else if(command=="STAT" ||
          command=="LIST" ||
          command=="UIDL" ||
          command=="RETR" ||
          command=="TOP" ||
          command=="DELE" ||
          command=="LAST" ||
          command=="RSET")
    out << "-ERR please login first\r\n";
  else
    out << "-ERR Invalid command: \"" << command << "\"\r\n";

  out.flush();

  return false;
 }

/*******************************************************************\

Function: dbpop3dt::do_command_authenticated

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbpop3dt::do_command_authenticated()
 {
  if(command=="QUIT")
   {
    mailfolder.lock("spool_in_folder");

    bool crunch=false;

    for(unsigned i=0; i<messages.size(); i++)
      if(deleted[i])
       {
        dword nr=messages[i];
        lseek(index, nr*sizeof(fi), SEEK_SET);
        read(index, &fi, sizeof fi);

        if(!fi.flags.deleted)
         {
          fi.flags.deleted=true;
          lseek(index, nr*sizeof(fi), SEEK_SET);
          write(index, &fi, sizeof fi);
          crunch=true;
         }
       }

    mailfolder.unlock();

    #ifndef DBOX_STANDALONE // TODO
    if(crunch)
      docrunch(user);
    #endif

    out << "+OK " << client.config["hostname"]
        << " DBOX POP Server (Version " DBOXVERSIONNR ") shutdown.\r\n";

    return true;
   }
  else if(command=="USER")
   {
    out << "-ERR already authenticated\r\n";
   }
  else if(command=="PASS")
   {
    out << "-ERR already authenticated\r\n";
   }
  else if(command=="STAT")
   {
    if(index<0)
      out << "+OK 0 0\r\n";
    else
     {
      dword total_size=0; // estimated

      for(dword arg=1; arg<=messages.size(); arg++)
       {
        if(!deleted[arg-1])
         {
          get_folderindex(messages[arg-1]);

          if(!fi.flags.binary)
            total_size+=fi.len+fi.headerlen;
          else
            total_size+=fi.headerlen+fi.len*1377/1000; // uuencode
         }
       }

      out << "+OK " << messages.size() << " " 
                          << total_size << "\r\n";
     }
   }
  else if(command=="LIST")
   {
    if(args=="")
     {
      out << "+OK " << messages.size() << " messages\r\n";

      for(dword arg=1; arg<=messages.size(); arg++)
        dolist(arg, false, false);

      out << ".\r\n";
     }
    else
     {
      dword arg=atoi(args.c_str());
      if(arg==0 || arg>messages.size())
        out << "-ERR invalid message\r\n";
      else
        dolist(arg, true, false);
     }
   }
  else if(command=="UIDL")
   {
    if(args=="")
     {
      out << "+OK " << messages.size() << " messages\r\n";

      for(dword arg=1; arg<=messages.size(); arg++)
        dolist(arg, false, true);

      out << ".\r\n";
     }
    else
     {
      dword arg=atoi(args.c_str());
      if(arg==0 || arg>messages.size())
        out << "-ERR invalid message\r\n";
      else
        dolist(arg, true, true);
     }
   }
  else if(command=="RETR" || command=="TOP")
   {
    const char *toparg=(char *)NULL;

    if(command=="TOP")
     {
      toparg=strrchr(args.c_str(), ' ');
      if(toparg!=NULL) toparg++;
     }

    if(args=="")
      out << "-ERR message number required!\r\n";
    else
     {
      dword arg, nr;

      arg=atoi(args.c_str());
      if(arg==0 || arg>messages.size())
        out << "-ERR invalid message\r\n";
      else if(deleted[arg-1])
        out << "-ERR message is deleted\r\n";
      else
       {
        if(arg>lastmsg) lastmsg=arg;
        nr=messages[arg-1];
        retr_top(nr, toparg);
       }
     }
   }
  else if(command=="DELE")
   {
    if(args=="")
      out << "-ERR message number required!\r\n";
    else
     {
      dword arg=atoi(args.c_str());

      if(arg==0 || arg>messages.size())
        out << "-ERR invalid message\r\n";
      else if(deleted[arg-1])
        out << "-ERR message is deleted\r\n";
      else
       {
        if(arg>lastmsg) lastmsg=arg;
        deleted[arg-1]=true;
        out << "+OK message deleted.\r\n";
       }
     }
   }
  else if(command=="NOOP")
   {
    if(args!="")
      out << "-ERR Uh?\r\n";
    else
      out << "+OK Gaehn.\r\n";
   }
  else if(command=="APOP")
   {
    out << "-ERR sorry, no APOP supported\r\n";
   }
  else if(command=="LAST")
   {
    if(args!="")
      out << "-ERR Uh?\r\n";
    else
      out << "+OK " << lastmsg << "\r\n";
   }
  else if(command=="RSET")
   {
    if(args!="")
      out << "-ERR Uh?\r\n";
    else
     {
      for(unsigned i=0; i<deleted.size(); i++)
        deleted[i]=false;

      out << "+OK\r\n";
     }
   }
  else
    out << "-ERR Invalid command: \"" << command << "\"\r\n";

  out.flush();

  return false;
 }

/*******************************************************************\

Function: dopop3d

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbpop3d()
 {
  dbpop3dt dbpop3d(std::cin, std::cout);

  dbpop3d.dbpop3d();
 }
