/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: base class for remote mailbox access

*/

#include <dblib/os_dep_file.h>

#include <dblib/getpeer.h>
#include <dblib/str_getline.h>
#include <dbclient/logstr.h>
#include <dbclient/auth.h>
#include <dbclient/client.h>
#include <dbsmtpd/folder_path.h>
#include "pop.h"

/*******************************************************************\

Function: popt::popt

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

popt::popt(std::istream &_in, std::ostream &_out):
           in(_in), out(_out)
 {
  protocol="unknown";
  timeout_minutes=10;
  authenticated=false;
 }

/*******************************************************************\

Function: popt::authenticate

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool popt::authenticate(const std::string &password)
 {
  try
   {
    autht auth;
    auth.auth(user, password);
   }

  catch(auth_excpt e)
   {
    logstr << "authentication of user " << user
           << " failed: " << e.what();
    logstr.log("popt::authenticate");
    return true;
   }

  authenticated=true;

  return false;
 }

/*******************************************************************\

Function: popt::open_mail

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool popt::open_mail(const std::string &folder_name)
 {
  mailfolder.set_folder_name(folder_name);
  mailfolder.set_user(user);

  return open_mail();
 }

/*******************************************************************\

Function: popt::open_mail

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool popt::open_mail()
 {
  return open(mailfolder.get_path());
 }

/*******************************************************************\

Function: popt::getcommand

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool popt::getcommand()
 {
  bool got_timeout;

  if(!str_getline(in, commandline, true, timeout_minutes*60,
                  got_timeout))
    return true;

  if(got_timeout)
   {
    timeout();
    return true;
   }

  return false;
 }

/*******************************************************************\

Function: popt::getpeer

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void popt::getpeer()
 {
  peer=::getpeer(0);
 }

/*******************************************************************\

Function: popt::docrunch

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void popt::docrunch(const std::string &username)
 {
  #if 0
  char filename[DIRLEN];
  FILE *outfile;

  snprintf(filename, DIRLEN, BOXTODODIR "crunch.%s", config.boxdir, username);
  outfile=fopen(filename, "wt");
  if(outfile!=NULL) fclose(outfile);
  #endif
 }

/*******************************************************************\

Function: popt::movemail

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

#ifdef DBOX_STANDALONE
void dboxpopt::movemail()
 {
  ::movemail(up);
 }
#endif
