/*

  DBOX Module
  Copyright (C) 1994-2003 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <libdbox/dboxversion.h>
#include <dblib/str_getline.h>
#include <dblib/rfc1123date.h>
#include <dblib/getpeer.h>
#include <dblib/os_dep_file.h>
#include <dblib/os_dep_process.h>
#include <dbclient/client.h>
#include <dbclient/logstr.h>

#include "deliver.h"
#include "dbsmtpd_class.h"

/*******************************************************************\

Function: dbsmtpdt::strip_brackets

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

std::string dbsmtpdt::strip_brackets(const std::string &s)
 {
  if(!s.empty() && s[0]=='<' && s[s.size()-1]=='>')
    return std::string(s, 1, s.size()-2);

  return s;
 }

/*******************************************************************\

Function: dbsmtpdt::strip_space

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

std::string dbsmtpdt::strip_space(const std::string &s)
 {
  unsigned i=0;

  while(s.size()>i && s[i]==' ') i++;

  return std::string(s, i);
 }

/*******************************************************************\

Function: dbsmtpdt::parse_command

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::parse_command()
 {
  if(command=="")
   {
    arg="";
    return;
   }

  std::string::size_type pos=command.find(' ');
  if(pos==std::string::npos)
    return;

  arg=std::string(command, pos+1);
  command=std::string(command, 0, pos);
 }

/*******************************************************************\

Function: dbsmtpdt::check_rcpt

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::check_rcpt(const std::string &address)
 {
  std::string error;

  rcptt r;
  r.rcpt=address;
  get_user(address, r.user, error);

  if(error=="")
   {
    status_response(250, "OK");
    rcpt.push_back(r);
   }
  else
   {
    status_response(550, error);
    logstr << "rcpt rejected: " << address
           << " (" << error << ")";
    logstr.log("check_rcpt");
   }
 }

/*******************************************************************\

Function: dbsmtpdt::vrfy

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::vrfy(const std::string &address)
 {
  std::string error, user;

  get_user(address, user, error);

  if(error=="")
    status_response(252, "OK");
  else
    status_response(550, error);
 }

/*******************************************************************\

Function: dbsmtpd::help

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::help()
 {
  out << "214-The following commands are accepted:\r\n"
         "214-helo noop mail data rcpt help quit rset expn vrfy\r\n"
         "214-\r\n"
         "214 Send bugs reports to: kroening@handshake.de\r\n";
 }

/*******************************************************************\

Function: dbsmtpdt::get_user

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::get_user(const std::string &address,
                        std::string &user,
                        std::string &error)
 {
  error=user="";

  // see if it is the postmaster

  if(address=="postmaster")
   {
    get_user(address+"@"+hostname, user, error);
    return;
   }

  // first a sanity check

  if(address.find(' ')!=std::string::npos)
   {
    error="invalid address";
    return;
   }

  std::string::size_type p=address.find('@');
  if(p==std::string::npos)
   {
    error="invalid address";
    return;
   }

  std::string host(address, p+1);

  // first try specific address

  if(!mail_index.find(address, user))
   {
    if(user=="")
      error="address invalid";
    else if(user=="localhost")
      get_user_localhost(address, user, error);

    return;
   }

  // try host

  if(host=="localhost")
   {
    get_user_localhost(address, user, error);
    return;
   }

  if(!mail_index.find(host, user))
   {
    if(user=="")
      error="host address invalid";
    else if(user=="localhost")
      get_user_localhost(address, user, error);

    return;
   }

  // try domains

  std::string domain=host;

  while(true)
   {
    if(!mail_index.find("."+domain, user))
     {
      if(user=="")
        error="host address invalid";
      else if(user=="localhost")
        get_user_localhost(address, user, error);

      return;
     }

    std::string::size_type s=domain.find('.');
    if(s==std::string::npos) break;
    domain=std::string(domain, s+1);
   }

  error="no service for address";
 }

/*******************************************************************\

Function: dbsmtpdt::get_user_localhost

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::get_user_localhost(const std::string &address,
                                  std::string &user,
                                  std::string &error)
 {
  error=user="";

  // first a sanity check

  if(address.find(' ')!=std::string::npos)
   {
    error="invalid address";
    return;
   }

  std::string::size_type p=address.find('@');
  if(p==std::string::npos)
   {
    error="invalid address";
    return;
   }

  std::string username(address, 0, p);

  if(!mail_index.find(username+"@localhost", user))
   {
    if(user=="")
     {
      error="user account "+username+" invalid";
      return;
     }
    
    // see if it is actually there
    std::string user_dir=build_path(maildir, user);

    if(access(user_dir.c_str(), F_OK)!=0)
     {
      error="directory of user "+user+" not available";
      return;
     }

    // OK

    return;
   }

  error="user account "+username+" not found";
 }

/*******************************************************************\

Function: dbsmtpdt::check_mail

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::check_mail(const std::string &address)
 {
  // first a sanity check

  if(address=="")
   {
   }
  else
   {
    if(address.find(' ')!=std::string::npos)
     {
      status_response(553, "no space allowed");
      return;
     }

    std::string::size_type p=address.find('@');
    if(p==std::string::npos)
     {
      status_response(553, "domain part missing");
      return;
     }
   }

  status_response(250, "OK");
  mail_from=address;
  mail_from_set=true;
 }

/*******************************************************************\

Function: dbsmtpdt::dbsmtpdt

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

dbsmtpdt::dbsmtpdt(std::istream &_in, std::ostream &_out):
  in(_in), out(_out)
 {
 }

/*******************************************************************\

Function: dbsmtpdt::dbsmtpdt

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::dbsmtpd()
 {
  // set umask for safety
  umask(077);

  try
   {  
    hostname=client.config["hostname"];
    maildir=client.config["maildir"];
    std::string maildb=client.config["maildb"];
    peer=getpeer(0);

    if(maildb=="" || maildir=="" || hostname=="")
     {
      status_response(450, "Mail system configuration error");
      return;
     }

    try
     {
      mail_index.open(maildb);
     }

    catch(excpt e)
     {
      status_response(450, "Mail system database unavailable");
      return;
     }

    status_response(220, hostname+" SMTP dbsmtpd");

    mail_from_set=quit=false;

    do
     {
      read_command();

      if(quit)
        break;

      parse_command();

      process_request();
     }
    while(!quit);
   }

  catch(...)
   {
    status_response(450, "internal server error");
   }
 }

/*******************************************************************\

Function: dbsmtpdt::read_command 

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::read_command()
 {
  bool got_timeout;

  if(!str_getline(in, command, true, 600, got_timeout))
    quit=true;
  else if(got_timeout)
   {
    status_response(451, "timeout waiting for input");
    logstr << "timeout while talking to " << peer;
    logstr.log("read_command");
    quit=true;
   }
 }

/*******************************************************************\

Function: dbsmtpdt::process_request

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::process_request()
 {
  upstring(command);

  if(command=="QUIT")
   {
    status_response(221, "Bye");
    quit=true;
   }
  else if(command=="HELO")
   {
    helo=arg;
    status_response(250, hostname);
   }
  else if(command=="RSET")
   {
    mail_from_set=false;
    mail_from="";
    rcpt.clear();
    status_response(250, "OK");
   }
  else if(command=="HELP")
   {
    help();
   }
  else if(command=="NOOP")
   {
    status_response(250, "OK");
   }
  else if(command=="RCPT")
   {
    if(!mail_from_set)
       status_response(503, "need MAIL command");
    else if(up_string(std::string(arg, 0, 3))=="TO:")
      check_rcpt(preprocess(std::string(arg, 3)));
    else
      status_response(501, "Syntax: RCPT TO:<address>");
   }
  else if(command=="VRFY")
    vrfy(preprocess(arg));
  else if(command=="MAIL")
   {
    if(up_string(std::string(arg, 0, 5))=="FROM:")
      check_mail(preprocess(std::string(arg, 5)));
    else
      status_response(501, "Syntax: MAIL FROM:<address>");
   }
  else if(command=="DATA")
   {
    if(rcpt.size()==0)
      status_response(503, "Error: need RCPT");
    else
      get_data();
   }
  else
    status_response(502, "Error: command not implemented");
 }

/*******************************************************************\

Function: dbsmtpdt::status_response

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::status_response(unsigned code, const std::string &msg)
 {
  out << code << " " << msg << "\r\n";
 }

/*******************************************************************\

Function: dbsmtpd::add_received_header

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::add_received_header(std::list<std::string> &lines)
 {
  std::string received;

  received="Received: from ";
  received+=helo;
  received+=" (["+peer+"])";
  lines.push_back(received);
  received="          by "+hostname;
  received+=" (" DBOXVERSIONNR ")";
  received+="; ";
  received+=rfc1123date(time(NULL), 0);
  lines.push_back(received);
 }

/*******************************************************************\

Function: dbsmtpd::add_message_id_header

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::add_message_id_header(std::list<std::string> &lines)
 {
  unsigned pid;

  #ifdef WIN32
  pid=0;
  #else
  pid=getpid();
  #endif

  std::string message_id;

  static unsigned int midcount=0;

  message_id="Message-ID: ";
  message_id+="<dbsmtpd."+
              i2string((pid+(midcount++)))+"."+
              i2string((time(NULL)))+"@"+
              hostname+">";
  lines.push_front(message_id);
 }

/*******************************************************************\

Function: dbsmtpdt::get_data

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::get_data()
 {
  status_response(354, "End data with <CR><LF>.<CR><LF>");

  std::string line;
  std::list<std::string> lines;

  add_received_header(lines);

  bool in_header=true,
       message_id_seen=false;

  while(true)
   {
    bool got_timeout;

    if(!str_getline(in, line, true, 600, got_timeout))
     {
      quit=true;
      logstr << "connection lost while talking to " << peer;
      logstr.log("get_data");
      return;
     }
    else if(got_timeout)
     {
      status_response(451, "timeout waiting for input");
      quit=true;
      logstr << "timeout while talking to " << peer;
      logstr.log("get_data");
      return;
     }

    if(line==".")
      break;
    else if(line.size()!=0 && line[0]=='.')
      line.erase(0, 1);
    else if(line=="")
      in_header=false;

    lines.push_back(line);

    if(in_header)
     {
      std::string::size_type pos=line.find_first_of(" :\t");

      if(pos!=std::string::npos)
       {
        std::string header(line, 0, pos);
        dnstring(header);

        if(header=="message-id")
          message_id_seen=true;
       }
     }
   }

  if(!message_id_seen)
    add_message_id_header(lines);

  deliver(lines);

  rcpt.clear();
  mail_from_set=false;
 }

/*******************************************************************\

Function: dbsmtpdt::deliver_puffer_erfc

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::deliver_puffer_erfc(const rfcmessaget &message,
                                   const std::string &user)
 {
  // found maps file

  std::string puffer_erfc=
    build_path(build_path(maildir, user), user+".erfc");

  std::ofstream out(puffer_erfc.c_str(), 
                    std::ios::out|std::ios::binary|std::ios::app);

  if(!out)
   {
    logstr << "failed to append mail to " << puffer_erfc;
    logstr.log("deliver_puffer_erfc");
   }
  else
   {
    for(std::list<rcptt>::const_iterator it=rcpt.begin();
        it!=rcpt.end(); it++)
      if(it->user==user)
        out << "RCPT: " << it->rcpt << "\r\n";

    out << "FROM: " << mail_from << "\r\n";

    out << "LEN: " << message.len+message.headerlen << "\r\n";
    out << "\r\n";
    message.copy(out);
   }
 }

/*******************************************************************\

Function: dbsmtpdt::deliver_puffer_erfc

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::deliver_puffer_erfc(const rfcmessaget &message)
 {
  // get set of users

  std::set<std::string> user_set;

  for(std::list<rcptt>::const_iterator it=rcpt.begin();
      it!=rcpt.end(); it++)
    user_set.insert(it->user);

  // deliver to set of users with maps file

  for(std::set<std::string>::const_iterator it=user_set.begin();
      it!=user_set.end(); it++)
   {
    if(access(build_path(build_path(maildir, *it), "maps").c_str(), 00)==0)
      deliver_puffer_erfc(message, *it);
   }
 }

/*******************************************************************\

Function: dbsmtpdt::deliver

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpdt::deliver(std::list<std::string> &lines)
 {
  // check for spam
  bool spam=is_spam(lines);

  // precompute size to prevent reallocation

  std::string::size_type s=0;

  for(std::list<std::string>::const_iterator it=lines.begin();
      it!=lines.end(); it++)
    s+=it->size()+2;

  std::string message_text;

  message_text.reserve(s);

  // add up lines

  for(std::list<std::string>::const_iterator it=lines.begin();
      it!=lines.end(); it++)
   {
    message_text+=*it;
    message_text+="\r\n";
   }

  // build message

  rfcmessaget message;
  message.data=message_text.c_str();
  message.endoffile=message.endofmail=message.data+message_text.size();

  // parse header
  message.parse();

   {
    std::string header;

    // log the message
    logstr << "from=<" << mail_from << ">";
    logstr << ", size=" << message_text.size();

    if(!message.findsingleheader(header, "Message-ID"))
      logstr << ", msgid=" << header;

    logstr << ", relay=" << peer;

    logstr << ", spam=" << spam;

    logstr.log("deliver");
   }

  // deliver message

  try
   {
    for(std::list<rcptt>::const_iterator it=rcpt.begin();
        it!=rcpt.end(); it++)
     {
      if(it->user.empty())
        throw excpt("internal error: empty user (1)");

      ::deliver(message, it->user, it->rcpt, mail_from, spam);

      // log it
      logstr << "to=<" << it->rcpt << ">";
      logstr << ", user=" << it->user;
      logstr.log("deliver");
     }

    status_response(250, "OK");

    // deliver to puffer.erfc - if required
    deliver_puffer_erfc(message);
   }

  catch(excpt e)
   {
    status_response(450, e.what());
    logstr << "delivery failed: " << e.what();
    logstr.log("deliver");
   }
 }

/*******************************************************************\

Function: dbsmtpd

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsmtpd()
 {
  dbsmtpdt dbsmtpd(std::cin, std::cout);

  dbsmtpd.dbsmtpd();
 }
