/*

  DBOX Module
  Copyright (C) 1994-2003 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <fstream>

#include <dblib/os_dep_file.h>
#include <dblib/os_dep_process.h>
#include <dblib/str_getline.h>
#include <dblib/os_dep_setjmp.h>
#include <dbclient/client.h>
#include <dbclient/logstr.h>

#include "dbsmtpd_class.h"
#include "folder_path.h"
#include "deliver.h"

/*******************************************************************\

Function: dbsmtpdt::check_spam

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbsmtpdt::check_spam()
 {
  std::set<std::string> user_set;

  for(std::list<rcptt>::const_iterator it=rcpt.begin();
      it!=rcpt.end(); it++)
    user_set.insert(it->user);

  for(std::set<std::string>::const_iterator it=user_set.begin();
      it!=user_set.end(); it++)
   {
    if(it->empty())
      throw excpt("internal error: empty user (1)");

    return folder_exists(maildir, *it, SPAM_FOLDER);
   }

  return false;
 }

/*******************************************************************\

Function: is_spam_pipe

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

jmp_buf is_spam_jmp;

void is_spam_pipe(int signal)
 {
  longjmp(is_spam_jmp, 1);
 }


/*******************************************************************\

Function: dbsmtpdt::is_spam

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool dbsmtpdt::is_spam(std::list<std::string> &lines)
 {
  const std::string &spamfilter=client.config["spamfilter"];

  if(spamfilter=="") return false;

  if(!check_spam()) return false;

  if(spamfilter=="spamassassin")
   {
    std::string temp_dir="/tmp/dbsmtpd."+i2string(getpid());
    if(mkdir(temp_dir.c_str(), 0700)!=0)
     {
      logstr << "mkdir(\"" << temp_dir << "\", 0700) failed";
      logstr.log("dbsmtpd::is_spam");
      return false;
     }

    std::string temp_file=build_path(temp_dir, "spam.result");

    std::string command="spamc -c -f > "+temp_file;
    bool result=false;
    FILE *spamc=NULL;

    if(setjmp(is_spam_jmp)!=0)
     {
      logstr << "spamc closed the pipe";
      logstr.log("dbsmtpd::is_spam");
     }
    else
     {
      signal(SIGPIPE, is_spam_pipe);

      spamc=popen(command.c_str(), "w");

      if(spamc==NULL)
       {
        logstr << "popen(\"" << command << "\") failed";
        logstr.log("dbsmtpd::is_spam");
       }
      else
       {
        for(std::list<std::string>::const_iterator
            it=lines.begin();
            it!=lines.end();
            it++)
         {
          fputs(it->c_str(), spamc);
          fputs("\n", spamc);
         }

        pclose(spamc);
       }

      signal(SIGPIPE, SIG_DFL);
     }

    std::ifstream in(temp_file.c_str());
    std::string line;

    if(in)
     {
      str_getline(in, line);

      result=atoi(line.c_str())!=0;

      in.close();
     }

    // clean up again
    unlink(temp_file.c_str());
    rmdir(temp_dir.c_str());

    if(result)
     {
      lines.push_front("X-Spam-Flag: YES");
      lines.push_front("X-Spam-Result: "+line);
     }

    return result;
   }

  return false;
 }

