/*

  DBOX Module
  Copyright (C) 1994-2003 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <time.h>

#include <fstream>

#include <dblib/os_dep_file.h>
#include <dblib/rfc1522.h>
#include <dblib/parse_from.h>
#include <libdbox/folderindex.h>
#include <libdbox/stringtools.h>

#include "folder_insert.h"
#include "extern_filename.h"
#include "folder_profile.h"

#define EXTERNLEN 4096

/*******************************************************************\

Function: parsedate

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

extern "C" time_t parsedate(char *line, int *tz);

static time_t parsedate(const std::string &header)
 {
  int tz;

  return parsedate((char *)header.c_str(), &tz);
 }

/*******************************************************************\

Function: folder_insert

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void folder_insert(rfcmessaget &m,
                   const std::string &directory,
                   bool forceextern,
                   const std::string &envelope_to,
                   const std::string &envelope_from,
                   dword &nr)
 {
  folderindext fi;

  std::string envelope_to_header;
  std::string envelope_from_header;

  if(envelope_to!="")
    envelope_to_header=
      "X-POP3-Rcpt:\t"+envelope_to+"\r\n"
      "Envelope-To:\t"+envelope_to+"\r\n";

  if(envelope_from!="")
    envelope_from_header="X-POP3-From:\t"+envelope_from+"\r\n";

  if(access(build_path(directory, ".").c_str(), 00)!=0)
    throw excpt("folder_insert failed to access "+directory);

  folder_profilet fp(directory);

   {
    std::string header_filename=build_path(directory, "header");

    std::ofstream outfile(header_filename.c_str(),
                          std::ios::out|std::ios::app|std::ios::binary);
    if(!outfile)
      throw excpt("folder_insert failed to open "+header_filename);

    fi.clear();

    std::string header; 

    m.findsingleheader(header, "Subject");
    header=decode_rfc1522(header);
    iso2ibm(header);
    strmaxcpy(fi.bet, header.c_str(), TEXTLEN-1);

     {
      std::string mailto, name;
      m.findsingleheader(header, "From");
      parse_from(header, mailto, name);
      std::string parsed_from=mailto;
      if(name!="") parsed_from+=" ("+decode_rfc1522(name)+")";
      strmaxcpy(fi.abs, parsed_from.c_str(), TEXTLEN-1);
     }

    m.findsingleheader(header, "Message-ID"); // TODO: mehr parsen
    strmaxcpy(fi.mid, header.c_str(), TEXTLEN-1);

    m.findsingleheader(header, "Date"); // TODO: mehr parsen
    fi.eda=parsedate(header);

    time(&fi.received);

    fi.len=m.len;
    fi.headerlen=m.headerlen+envelope_from_header.size()+
                             envelope_to_header.size();
    fi.flags.bodyextern=(forceextern || fi.len>EXTERNLEN);
    fi.flags.binary=FALSE;
    fi.flags.rfc=TRUE;
    fi.headeroffs=filelength(header_filename);
    fi.id=atol(fp["nextid"].c_str());
    if(fi.id==0) fi.id=1;

    outfile << envelope_to_header << envelope_from_header;

    if(fi.flags.bodyextern)
     {
      if(m.copyheader(outfile))
        throw excpt("folder_insert failed to copy header");

      std::string externalfilename, hint, prefix;
      m.findsingleheader(hint, "Subject");
      prefix="msg"+i2string(fi.id)+"_";
      extern_filename(externalfilename, prefix, hint, directory);

      std::ofstream outfile2(externalfilename.c_str(), std::ios::out|std::ios::binary);
      if(!outfile2)
        throw excpt("folder_insert failed to open "+externalfilename);

      strmaxcpy(fi.file, strrchr(externalfilename.c_str(), 
        DIRCHAR[0])+1, TEXTLEN-1);

      m.copybody(outfile2);
     }
    else
     {
      if(m.copy(outfile))
        throw excpt("folder_insert failed to copy message");
     }
   }

   {
    std::string index_filename=build_path(directory, "index");

    std::ofstream indexfile(index_filename.c_str(),
                            std::ios::out|std::ios::app);

    if(!indexfile)
      throw excpt("folder_insert failed to open "+index_filename);

    nr=indexfile.tellp()/sizeof fi;

    if(!indexfile.write((char *)&fi, sizeof fi))
      throw excpt("folder_insert failed to write to "+index_filename);
   }

  fp.set("nextid", fi.id+1);
  fp.write();
 }
