/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <signal.h>
#include <fcntl.h>
#include <unistd.h>

#include <iostream>
#include <fstream>

#include <dblib/table.h>
#include <dblib/str_getline.h>

#include <libdbox/stringtools.h>
#include <libdbox/dboxversion.h> 
#include <libdbox/flags.h>
#include <libdbox/dboxpath.h>
#include <libdbox/userdb.h>
#include <libdbox/logfile.h>
#include <libdbox/log_serv.h>
#include <libdbox/lock.h>

#include "dbsrv.h"

char alarm_lockfile_type[TEXTLEN]="",
     alarm_lockfile_name[TEXTLEN]="";

/*******************************************************************\

Function: dbsrvalarm

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsrvalarm(int signal)
 {
  std::cout << "-ERR logout, timeout\r\n";

  if(alarm_lockfile_type[0]!=0)
    unlock(alarm_lockfile_type, alarm_lockfile_name, config.boxname);

  log_serv.printf("dbsrvalarm", "aborting due to signal");

  exit(0);
 }

/*******************************************************************\

Function: dbsrvt::parse

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsrvt::parse()
 {
  std::string component;

  id.clear();

  for(unsigned i=0; i<name.size(); i++)
   {
    if(name[i]=='/')
     {
      id.push_back(component);
      component="";
     }
    else
      component+=name[i];
   }

  id.push_back(component);
 }

/*******************************************************************\

Function: dbsrvt::dodbsrv

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsrvt::dodbsrv()
 {
  auth_ok=FALSE;
 
  fcntl(fileno(stdout), F_SETFL, 0);

  std::cout << "+OK " << config.boxname << "." << config.domain
            << " DBOX dbsrv " DBOXVERSION " ready\r\n";

  getpeer();

  signal(SIGALRM, dbsrvalarm);
  signal(SIGPIPE, dbsrvalarm);

  while(TRUE)
   {
    std::string commandline;

    fflush(stdout);
    alarm(10*60);
    if(!str_getline(in, commandline)) return;
    alarm(0);

    const char *tptr=commandline.c_str();
    while(*tptr==' ' || *tptr=='\t') tptr++;

    std::vector<std::string> args;

    while(*tptr!=0)
     {
      std::string arg;

      while(*tptr!=0 && *tptr!=' ' && *tptr!='\t')
        arg+=*(tptr++);

      args.push_back(arg);

      while(*tptr==' ' || *tptr=='\t') tptr++;
     }

    std::string command;

    if(args.size()!=0) command=args[0];
    dnstring(command);
    
    if(command=="quit")
     {
      std::cout << "+OK DBOX dbsrv " << DBOXVERSION << " shutdown.\r\n";
      return;
     }     
    else if(command=="auth")
     {
      if(args.size()!=3)
        std::cout << "-ERR this command takes two arguments\r\n";
      else
        auth(args[1], args[2]);
     }
    else if(command=="get")
     {
      if(args.size()!=2)
        std::cout << "-ERR this command takes one argument\r\n";
      else
       {
        name=args[1];
        parse();
        get();
       }
     }
    else if(command=="service")
     {
      if(args.size()!=2)
        std::cout << "-ERR this command takes one argument\r\n";
      else
       {
        service=args[1];
        std::cout << "+OK\r\n";
       }
     }
    else
      std::cout << "-ERR command " << command << " unknown\r\n";
   }
 }

/*******************************************************************\

Function: dbsrvt::getpeer

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dbsrvt::getpeer()
 {
  struct sockaddr_in sin;
  unsigned int sval;

  sval=sizeof(sin);
  if(getpeername(0, (struct sockaddr *)&sin, &sval)<0)
   {
    peer=0;
    return;
   }

  peer=sin.sin_addr.s_addr;
 }

/*******************************************************************\

Function: dodbsrv

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void dodbsrv(std::istream &in, std::ostream &out)
 {
  dbsrvt dbsrv(in, out);

  dbsrv.dodbsrv();
 }
