/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: Convert DM to EURO

*/

#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <dbox.h>
#include <tools.h>
#include <charsettools.h>
#include <dboxversion.h>
#include <libdboxversion.h>
#include <userdb.h>
#include <account.h>
#include "euro.h"

void convert(amountt &amount)
 {
  // eur = DM / 1.95583

  unsigned old_scale=amount.scale;

  long double a=((long double)(amount.amount))/power10(amount.scale);

  a=a/double(1.95583);

  amount.amount=(long int)(a*power10(3));
  amount.scale=3;

  if(old_scale==2)
   {
    bool round_up=(amount.amount%10)>=5;
    amount.amount=amount.amount/10;
    amount.scale=2;
    if(round_up) amount.amount++;
   }
 }

void doeuro(userpt &up)
 {
  std::string oldname, newname;
  oldname=std::string(up.getpath())+"konto";
  newname=oldname+".dm";

  if(access(oldname.c_str(), 0)!=0)
    return;

  if(access(newname.c_str(), 0)==0)
   {
    std::cerr << "user " << up.name << " already is converted" << std::endl;
    abort();
   }

  if(up.lock("euro"))
   {
    std::cerr << "failed to lock user " << up.name << std::endl;
    abort();
   }

  if(rename(oldname.c_str(), newname.c_str())!=0)
   {
    std::cerr << "rename for user " << up.name << " failed" << std::endl;
    abort();
   }

  up.set("auszug.1.uebertrag.scale", "");
  up.set("auszug.1.uebertrag", "");
  up.set("auszug.1.buchungsnummer", "");
  up.set("auszug.2.uebertrag.scale", "");
  up.set("auszug.2.uebertrag", "");
  up.set("auszug.2.buchungsnummer", "");

  up.set("dm.buchungsnummer", up|"buchungsnummer");
  up.set("dm.konto", up|"konto");
  up.set("dm.konto.scale", up|"konto.scale");

  up.set("buchungsnummer", 0);
  amountt amount=getbalance(&up);
  amountt old_amount=amount;
  convert(amount);

  up.set("konto", 0);

  up.write();
  up.unlock();

  if(amount.amount!=0)
   {
    buchungt buchung;

    buchung.amount=amount;
    if(buchung.amount.amount>0)
      buchung.haben=TRUE;
    else
     {
      buchung.haben=FALSE;
      buchung.amount.amount=-buchung.amount.amount;
     }

    strmaxcpy(buchung.bearbeiter, "sysop", TEXTLEN-1);
    snprintf(buchung.verwendungszweck, TEXTLEN, "Uebernahme DM Konto ");
    currency_string(buchung.verwendungszweck+strlen(buchung.verwendungszweck),
                    old_amount, TRUE, SYMBOL_ASCII);
    snprintf(buchung.buchungstyp, TEXTLEN, "ZAHLUNG.DM2EURO");

    dobuchung(&up, &buchung);
   }

  bankpt bp(up.name);
  if(bp.lock())
    fprintf(stderr, "Failed to lock bankprofile: %s\n", up.name);
  else
   {
    bp.set("abbuchungs_betrag", atol(bp|"abbuchungs_betrag")/2);
    bp.write();
    bp.unlock();
   }
 }

void doeuro()
 {
  char username[TEXTLEN], *tptr;

  while(fgets(username, TEXTLEN-1, stdin)!=NULL)
    if(username[0]!=0 && username[0]!='.' && username[0]!='_')
     {
      tptr=strchr(username, '\r'); if(tptr!=NULL) *tptr=0;
      tptr=strchr(username, '\n'); if(tptr!=NULL) *tptr=0;

      username[TEXTLEN-1]=0;
      tptr=strrchr(username, '/');
      if(tptr!=NULL) *tptr=0;
      userpt up(username);
      if(up.read())
        fprintf(stderr, "Not found: %s\n", username);
      else
        doeuro(up);
     }
 }

