#include <dblib/os_dep_file.h>

#include <fstream>

#include "folder_access.h"

/*******************************************************************\

Function: folder_accesst::folder_accesst

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

folder_accesst::folder_accesst()
 {
  cached_nr=0xffffffff;
  rfcm.data=(char *)NULL;
  mimestruct=(mimestructt *)NULL;
  index=header=-1;
  opened=FALSE;
 }

/*******************************************************************\

Function: folder_accesst::~folder_accesst

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

folder_accesst::~folder_accesst()
 {
  clear_cached();
  if(index>=0) ::close(index);
  if(header>=0) ::close(header);
 }

/*******************************************************************\

Function: folder_accesst::get_number_of_messages

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

dword folder_accesst::get_number_of_messages()
 {
  if(index<0) return 0;

  return filelength(index)/sizeof fi;
 }

/*******************************************************************\

Function: folder_accesst::close

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool folder_accesst::close()
 {
  clear_cached();
  if(index>=0) ::close(index);
  if(header>=0) ::close(header);
  index=header=-1;
  opened=FALSE;
  return FALSE;
 }

/*******************************************************************\

Function: folder_accesst::open

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool folder_accesst::open(const std::string &__path)
 {
  _path=__path;
  return _open();
 }

/*******************************************************************\

Function: folder_accesst::_open

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool folder_accesst::_open()
 {
  std::string filename;

  if(opened) close();

  filename=build_path(_path, "index");
  index=::open(filename.c_str(), O_RDWR|O_BINARY|O_CREAT, S_IREAD|S_IWRITE);

  filename=build_path(_path, "header");
  header=::open(filename.c_str(), O_RDWR|O_BINARY|O_CREAT, S_IREAD|S_IWRITE);

  return index<0 || header<0;
 }

/*******************************************************************\

Function: folder_accesst::search_uid

  Inputs:

 Outputs:

 Purpose: Binary search for uids

\*******************************************************************/

bool folder_accesst::search_uid(dword uid, dword &nr)
 {
  if(index<0) return TRUE;

  dword number_of_messages=get_number_of_messages();

  #if 1
  dword begin=0, end=number_of_messages;

  while(begin<end)
   {
    nr=begin+(end-begin)/2;

    get_folderindex(nr);

    if(fi.id==uid)
      return FALSE;
    
    if(fi.id<uid)
      begin=nr+1;
    else // fi.uid>uid
      end=nr;
   }
  #else
  for(nr=0; nr<number_of_messages; nr++)
   {
    get_folderindex(nr);

    if(fi.id==uid)
      return FALSE;
   }
  #endif
  return TRUE;
 }

/*******************************************************************\

Function: folder_accesst::clear_cached

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void folder_accesst::clear_cached()
 {
  if(rfcm.data!=(char *)NULL) { delete (char *)rfcm.data; rfcm.data=(char *)NULL; }
  if(mimestruct!=(mimestructt *)NULL) { delete mimestruct; mimestruct=(mimestructt *)NULL; }
  cached_body=cached_folderindex=cached_header=cached_structure=FALSE;
 }

/*******************************************************************\

Function: folder_accesst::check_cache

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void folder_accesst::check_cache(dword nr)
 {
  if(nr!=cached_nr)
   {
    clear_cached();
    cached_nr=nr;
   }
 }

/*******************************************************************\

Function: folder_accesst::get_folderindex

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool folder_accesst::get_folderindex(dword nr)
 {
  check_cache(nr);

  if(!cached_folderindex) 
   {
    lseek(index, nr*sizeof(fi), SEEK_SET);
    if(read(index, &fi, sizeof fi)!=sizeof fi)
      return TRUE;

    cached_folderindex=TRUE;
   }

  return FALSE;
 }

/*******************************************************************\

Function: folder_accesst::get_header

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool folder_accesst::get_header(dword nr)
 {
  check_cache(nr);

  if(get_folderindex(nr))
    return TRUE;

  if(!cached_header)
   {
    char *data;

    data=new char[fi.headerlen];
    if(data==(char *)NULL)
      abort();

    lseek(header, fi.headeroffs, SEEK_SET);

    if(read(header, data, fi.headerlen)!=(ssize_t)fi.headerlen)
     {
      delete data;
      return TRUE;
     }

    if(fi.flags.rfc)
     {
      if(rfcm.data!=NULL) delete (char *)rfcm.data;
      rfcm.data=data;
      rfcm.endofmail=rfcm.endoffile=data+fi.headerlen;
      rfcm.parse();
     }
    else // Zconnect
     {
     }

    cached_header=TRUE;
   }

  return FALSE;
 }

/*******************************************************************\

Function: folder_accesst::get_body

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool folder_accesst::get_body(dword nr)
 {
  check_cache(nr);

  if(get_folderindex(nr))
    return TRUE;

  if(!cached_body)
   {
    char *data;

    data=new char[fi.len+fi.headerlen];
    if(data==(char *)NULL) abort();

    lseek(header, fi.headeroffs, SEEK_SET);

    if(!fi.flags.bodyextern)
     {
      dword total=fi.len+fi.headerlen;
      if(read(header, data, total)!=(signed)total)
        return TRUE;
     }
    else
     {
      if(read(header, data, fi.headerlen)!=(signed)fi.headerlen)
        return TRUE;

      std::string filename=build_path(_path, fi.file);
      std::ifstream bodyfile(filename.c_str(), std::ios::binary|std::ios::in);
      if(!bodyfile) return TRUE;
      if(!bodyfile.read(data+fi.headerlen, fi.len)) return TRUE;
     }

    if(fi.flags.rfc)
     {
      if(rfcm.data!=NULL) delete (char *)rfcm.data;
      rfcm.data=data;
      rfcm.endofmail=rfcm.endoffile=data+fi.len+fi.headerlen;
      rfcm.parse();
     }
    else // Zconnect
     {
     }

    cached_body=TRUE;
    cached_header=TRUE;
   }

  return FALSE;
 }

/*******************************************************************\

Function: folder_accesst::get_structure

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool folder_accesst::get_structure(dword nr)
 {
  check_cache(nr);
 
  if(get_body(nr))
    return TRUE;

  if(!cached_structure)
   { 
    mimestruct=new mimestructt;
    mimestruct->mimeparse(rfcm);
    cached_structure=TRUE;
   }

  return FALSE;
 }
