#include <time.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <dboxpath.h>
#include <config.h>
#include <logfile.h>
#include <log_serv.h>

#ifdef USE_SYSLOG
#include <syslog.h>
#endif

log_servt log_serv;

/*******************************************************************\

Function: log_servt::printf

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

int log_servt::printf(const char *module, const char *format, ...)
 {
  #ifndef USE_SYSLOG

  std::string filename;
  FILE *logfile;

  if(port==0)
   {
    if(logfilename!=NULL && *logfilename!="")
      filename=*logfilename;
    else if(config.dboxdir[0]!=0)
     {
      getdboxsystemdir(filename, config.dboxdir); 
      filename+="logfile";
     }
    else
      filename=DIRCHAR "logfile"; // last resort...
   }
  else
   {
    getdboxportdir(filename, config.dboxdir, port);
    filename+="logfile";
   }

  if((logfile=fopen(filename.c_str(), "at"))==NULL)
   {
    fprintf(stderr, "FATAL: failed to open %s\r\n",
            filename.c_str());
    return 0;
   }

  va_list arglist;
  timefprintf(logfile);
  fprintf(logfile, "%s: ", module);
  va_start(arglist, format);
  vfprintf(logfile, format, arglist);
  va_end(arglist);
  fprintf(logfile, "\n");

  fclose(logfile);

  #else

  char buffer[1000];
  va_list arglist;
  va_start(arglist, format);
  vsnprintf(buffer, 1000, format, arglist);
  va_end(arglist);

  syslog(LOG_NOTICE, "%s: %s", module.c_str(), buffer);

  #endif

  return 1;
 }

/*******************************************************************\

Function: log_servt::internalerror

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void log_servt::internalerror(const char *module, const char *error)
 {
  fflush(stdout);
  fprintf(stderr, "\r\nInternal error: %s: %s\r\n", module, error);
  log_serv.printf(module, "Internal error: %s", error);
  abort();
 }

/*******************************************************************\

Function: log_servt::~log_servt

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

log_servt::~log_servt()
 {
  #ifdef USE_SYSLOG
  if(syslog_opened)
    ::closelog();
  #else
  if(logfilename!=NULL) delete logfilename;
  #endif
 }

/*******************************************************************\

Function: log_servt::openlog

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void log_servt::openlog(const std::string &program)
 {
  #ifdef USE_SYSLOG
  ::openlog(program, LOG_PID, LOG_MAIL);
  syslog_opened=TRUE;
  #else
  std::string logfile;
  getboxlogdir(logfile, config.boxdir);
  logfile+=program;
  logfile+=".log";
  setlogfile(logfile);
  #endif
 }
