#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>
#include <signal.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <dirent.h>

#include <fstream>

#include <dboxpath.h>
#include <tools.h>
#include <flags.h>
#include <config.h>
#include <lock.h>
#include <userdb.h>
#include <log_serv.h>
#include <logfile.h>
#include <oldtools.h>
#include <dboxversion.h> 
#include <spoolp.h>
#include <dblib/stream_copy.h>
#include "netd.h"

char alarm_lockfile_type[TEXTLEN]="",
     alarm_lockfile_name[TEXTLEN]="";

void netdalarm(int signal)
 {
  printf("-ERR logout, timeout\r\n");

  if(alarm_lockfile_type[0]!=0)
    unlock(alarm_lockfile_type, alarm_lockfile_name, config.boxname);

  log_serv.printf("netdalarm", "aborting due to signal");

  exit(0);
 }

class netdt
 {
 public:
  bool auth_ok;
  internetaddresst peer;
  char userlogfile[DIRLEN];
  userpt up;

  void dotransfer(const char *args);
  void getpeer();
  void donetd();
 };

void netdt::dotransfer(const char *args)
 {
  stringlistt arglist;
  int i;
  bool get=FALSE, put=FALSE, mail=FALSE, news=FALSE, zconnect=FALSE,
       erfc=FALSE, list=FALSE, maps=FALSE, profiledata=FALSE;
  const char *type="?";
  dword size=0;
  char source[TEXTLEN]="", packer[TEXTLEN]="";

  if(!auth_ok)
   {
    printf("-ERR login first\r\n");
    return;
   }

  if(args==NULL || args[0]==0)
   {
    printf("-ERR give args\r\n");
    return;
   }

  arglist.parsestring(args);

  for(i=0; arglist.strings[i]!=NULL; i++)
   {
    if(stricmp(arglist.strings[i], "get")==0)
      get=TRUE;
    else if(stricmp(arglist.strings[i], "put")==0)
      put=TRUE;
    else if(stricmp(arglist.strings[i], "zconnect")==0)
      zconnect=TRUE;
    else if(stricmp(arglist.strings[i], "erfc")==0)
      erfc=TRUE;
    else if(stricmp(arglist.strings[i], "list")==0)
      list=TRUE;
    if(stricmp(arglist.strings[i], "mail")==0)
     {
      mail=TRUE;
      type="mail";
     }
    else if(stricmp(arglist.strings[i], "news")==0)
     {
      news=TRUE;
      type="news";
     }
    else if(stricmp(arglist.strings[i], "maps")==0)
     {
      maps=TRUE; 
      type="maps";
     }
    else if(stricmp(arglist.strings[i], "profile")==0)
     {
      profiledata=TRUE; 
      type="profile";
     }
    else if(stricmp(arglist.strings[i], "size")==0)
     {
      i++;
      if(arglist.strings[i]==NULL) break;
      size=atol(arglist.strings[i]);
     }
    else if(stricmp(arglist.strings[i], "source")==0)
     {
      i++;
      if(arglist.strings[i]==NULL) break;
      strmaxcpy(source, arglist.strings[i], TEXTLEN-1);
     }
    else if(stricmp(arglist.strings[i], "packer")==0)
     {
      i++;
      if(arglist.strings[i]==NULL) break;
      strmaxcpy(packer, arglist.strings[i], TEXTLEN-1);
     }
   }

  if((get && put) || (!get && !put))
   {
    printf("-ERR Please make a decision.\r\n");
    return;
   }

  if(!zconnect && !profiledata && !erfc)
   {
    printf("-ERR file format not yet supported\r\n");
    return;
   }

  if(put)
   {
    if(packer[0]!=0)
     {
      printf("-ERR packer not yet supported\r\n");
      return;
     }

    if((mail && news) || (!mail && !news))
     {
      printf("-ERR please split mail/news\r\n");
      return;
     }

    if(source[0]!=0)
     {
      char tempstr[TEXTLEN];
      up.get("netd.transfersuperuser", tempstr);
      if(atoi(tempstr)!=1)
       {
        printf("-ERR permission denied\r\n");
        return;
       }
     }
    else
      strmaxcpy(source, up.name, TEXTLEN-1);

    if(size==0)
     {
      printf("-ERR no size\r\n");
      return;
     }

    char outfilename[DIRLEN];
    snprintf(outfilename, DIRLEN, 
      DBOXTEMPDIR "netd.%d.puf", config.dboxdir, getpid());
    std::ofstream outfile(outfilename, std::ios::out|std::ios::binary|std::ios::app);
    if(!outfile)
     {
      printf("-ERR internal error\r\n");
      return;
     }

    printf("+OK send %lu bytes\r\n", size);
    fflush(stdout);
    
    if(filecopy(std::cin, outfile, size) ||
       outfile.tellp()!=std::streampos(size))
     {
      printf("-ERR transfer failed\r\n");
      outfile.close();
      unlink(outfilename);
      log_serv.printf("dotransfer", "Failed to transfer %lu bytes %s from %s",
        size, type, up.name);
      return;
     }

    printf("+OK received %lu bytes\r\n", size);
    outfile.close();

    log_serv.printf("dotransfer", "Transfered %lu bytes %s from %s for %s",
      size, type, up.name, source); 
 
    if(mail)
      spool_zc_puffer_mail(outfilename, source);
    else
      spool_zc_puffer_news(outfilename, source);
   }
  else if(get)
   {
    if(((mail && news) || (!mail && !news && !maps)) && !profiledata)
     {
      printf("-ERR please split mail/news\r\n");
      return;
     }

    if(list)
     {
      DIR *dir;
      dirent *ent;
      char tempstr[TEXTLEN];
      char filename[DIRLEN];

      up.get("netd.transfersuperuser", tempstr);
      if(atoi(tempstr)!=1)
       {
        printf("-ERR permission denied\r\n");
        return;
       }

      sprintf(filename, BOXUSERDIR ".", config.boxdir, ".");
      dir=opendir(filename);
      if(dir==NULL)
       {
        printf("-ERR internal error\r\n");
        return;
       }

      printf("+OK data follows\r\n");

      while((ent=readdir(dir))!=NULL)
       {
        if(ent->d_name[0]!='_' && ent->d_name[0]!='.')
         {
          if(maps)
            sprintf(filename, BOXUSERDIR "maps", config.boxdir, ent->d_name);
          else if(zconnect)
            sprintf(filename, BOXSPOOLDIR "%s.%s", 
              config.boxdir, ent->d_name, ent->d_name, mail?"prv":"brt");
          else // erfc
            sprintf(filename, BOXSPOOLDIR "%s.%s", 
              config.boxdir, ent->d_name, ent->d_name, "erfc");
         
          if(access(filename, 00)==0)
           {
            userpt list_up(ent->d_name); 

            dnstring(list_up.name);
            if(!list_up.read())
             {
              if(maps)
                printf("%s\r\n", ent->d_name);
              else
               {
                list_up.get("netd.transferskip", tempstr);
                if(atoi(tempstr)!=1 )
                  printf("%s\r\n", ent->d_name);
               }
             }
           }
         }
       }

      closedir(dir);

      printf(".\r\n");
     }
    else // !list
     {
      if(profiledata)
       {
        unsigned i;

        userpt up2(source);

        if(source[0]==0 || up2.read())
          printf("-ERR no such user \"%s\"\r\n", source);
        else
         {
          printf("+OK data follows\r\n");

          for(i=0; i<up2.size(); i++)
           {
            if(strnicmp(up2.data[i].id, "userflag.", 9)==0 ||
               stricmp(up2.data[i].id, "nachname")==0 ||
               stricmp(up2.data[i].id, "vorname")==0 ||
               stricmp(up2.data[i].id, "telefon")==0 ||
               stricmp(up2.data[i].id, "allemps")==0 ||
               stricmp(up2.data[i].id, "system")==0 ||
               stricmp(up2.data[i].id, "domains")==0 ||
               stricmp(up2.data[i].id, "maps")==0 ||
               stricmp(up2.data[i].id, "firma")==0 ||
               stricmp(up2.data[i].id, "gruppe")==0)
              printf("%s=%s\r\n", up2.data[i].id, up2.data[i].data);
           }

          printf("\r\n");
         }
       }
      else
       {
        char filename[DIRLEN], newfilename[DIRLEN], response[TEXTLEN];
        int infd;

        if(source[0]==0)
          strcpy(source, up.name);

        if(maps)
          sprintf(newfilename, BOXUSERDIR "maps", 
            config.boxdir, source);
        else
         {
          if(zconnect)
           {
            sprintf(filename, BOXSPOOLDIR "%s.%s", 
              config.boxdir, source, source, mail?"prv":"brt");

            sprintf(newfilename, BOXSPOOLDIR "%s.%s.transfer", 
              config.boxdir, source, source, mail?"prv":"brt");

            strcpy(alarm_lockfile_type,
              mail?"puffer.prv.transfer":"puffer.brt.transfer");
           }
          else // erfc
           {
            sprintf(filename, BOXSPOOLDIR "%s.%s", 
              config.boxdir, source, source, "erfc");

            sprintf(newfilename, BOXSPOOLDIR "%s.%s.transfer", 
              config.boxdir, source, source, "erfc");

            strcpy(alarm_lockfile_type, "puffer.erfc.transfer");
           }
        
          strcpy(alarm_lockfile_name, source);

          if(lock(alarm_lockfile_type, alarm_lockfile_name, config.boxname, "dotransfer"))
           {
            printf("-ERR transfer file is locked\r\n");
            alarm_lockfile_type[0]=0;
            return;
           }

          if(access(newfilename, 00)!=0)
           {
            if(access(filename, 00)!=0)
             {
              printf("-ERR no data\r\n");
              unlock(alarm_lockfile_type, alarm_lockfile_name, config.boxname);
              alarm_lockfile_type[0]=0;
              return;
             }

            const char *lock_name;
            if(zconnect)
              lock_name=mail?"puffer.prv":"puffer.brt";
            else
              lock_name="puffer.erfc";

            if(lock(lock_name, source, config.boxname, "dotransfer"))
             {
              printf("-ERR %s is locked...\r\n", source);
              unlock(alarm_lockfile_type, alarm_lockfile_name, config.boxname);
              alarm_lockfile_type[0]=0;
              return;
             }

            if(rename(filename, newfilename)!=0)
             {
              printf("-ERR internal error\r\n");
              log_serv.printf("dotransfer", "rename \"%s\"->\"%s\" failed",
                filename, newfilename);
              unlock(lock_name, source, config.boxname);
              unlock(alarm_lockfile_type, alarm_lockfile_name, config.boxname);
              alarm_lockfile_type[0]=0;
              return;
             }

            unlock(lock_name, source, config.boxname);
           }
         }

        if((infd=open(newfilename, O_RDONLY))<0)
         {
          printf("-ERR no data\r\n");
          return;
         }

        size=filelength(infd);
        printf("+OK %ld bytes\r\n", size);
        fflush(stdout);

        filecopy(infd, fileno(stdout), size);
        fflush(stdout);  
        close(infd);
            
        alarm(10*60);
        if(fgets(response, 900, stdin)==NULL) response[0]=0;
        alarm(0);

        if(response[0]=='+')
         {
          if(mail || news) 
            unlink(newfilename);

          log_serv.printf("dotransfer", "transfered %ld bytes %s to %s for %s",
            size, type, up.name, source);
         }
        else
          log_serv.printf("dotransfer", "failed to transfer %ld bytes %s to %s for %s",
            size, type, up.name, source);

        if(mail || news)
         {
          unlock(alarm_lockfile_type, alarm_lockfile_name, config.boxname);
          alarm_lockfile_type[0]=0;
         }
       }
     }
   }
 }

void netdt::donetd()
 {
  char commandline[1000], *tptr;

  auth_ok=FALSE;
  up.name[0]=0;
 
  fcntl(fileno(stdout), F_SETFL, 0);

  printf("+OK %s.%s DBOX netd " DBOXVERSION " ready\r\n", 
    config.boxname, config.domain);

  fflush(stdout);
  getpeer();

  signal(SIGALRM, netdalarm);
  signal(SIGPIPE, netdalarm);

  while(TRUE)
   {
    fflush(stdout);
    alarm(10*60);
    if(fgets(commandline, 900, stdin)==NULL) return;
    alarm(0);

    tptr=strchr(commandline, '\r'); if(tptr!=NULL) *tptr=0;
    tptr=strchr(commandline, '\n'); if(tptr!=NULL) *tptr=0;

    tptr=commandline;
    while(*tptr!=0 && *tptr!=' ' && *tptr!='\t') tptr++;

    if(*tptr!=0)
     {
      *tptr=0;
      tptr++;
      while(*tptr==' ' || *tptr=='\t') tptr++;
     }

    if(stricmp(commandline, "QUIT")==0)
     {
      printf("+OK DBOX netd " DBOXVERSION " shutdown.\r\n");
      return;
     }     
    else if(stricmp(commandline, "USER")==0 && !auth_ok)
     {
      strmaxcpy(up.name, tptr, TEXTLEN-1);
      if(up.name[0]==0)
        printf("-ERR invalid username\r\n");
      else
       {
        dnstring(up.name);
        printf("+OK please send PASS command\r\n");
        sprintf(userlogfile, BOXUSERDIR "logfile", config.boxdir, up.name);
       }
     }
    else if(stricmp(commandline, "PASS")==0 && !auth_ok)
     {
      if(up.name[0]==0)
        printf("-ERR please provide an username first\r\n");
      else
       {
        if(up.read())
         {
          printf("-ERR invalid password, bye\r\n");
          return;
         }

        if(verifylogin(&up, tptr))
         {
          printf("-ERR invalid password, bye\r\n");
          logfileprintf(userlogfile, "netd", "Falsches Passwort!");
          return;
         }

         {
          grouppt gp(up|"gruppe");

          gp.read();

          if(!getinternalflag(&up, &gp, LOGIN_NETZ))
           {
            printf("-ERR access denied, bye\r\n");
            logfileprintf(userlogfile, "netd", "Zugriff verweigert!");
            return;
           }
         }

        printf("+OK Welcome, %s\r\n", up.name);

        auth_ok=TRUE;    
       }
     }
    else if(stricmp(commandline, "TRANSFER")==0)
      dotransfer(tptr);
    else
      printf("-ERR Kommando unbekannt\r\n");
   }
 }

void netdt::getpeer()
 {
  struct sockaddr_in sin;
  unsigned int sval;

  sval=sizeof(sin);
  if(getpeername(0, (struct sockaddr *)&sin, &sval)<0)
   {
    peer=0;
    return;
   }

  peer=sin.sin_addr.s_addr;
 }

void donetd()
 {
  netdt netd;

  netd.donetd();
 }
