
/*

    File: smtp/auth.c
    
    Copyright (C) 2001 by Daniel Kroening <kroening@handshake.de>

    This source is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This source is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <signal.h>
#include <syslog.h>

#include "smtp.h"
#include "auth.h"
#include "lib.h"
#include "ip-lib.h"

static unsigned char *base64_alphabet =
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

static int unbase64(ch) int ch; {
  int i;
  if (ch == '=') return 0;
  for (i = 0; i < 64; i++)
    if (ch == base64_alphabet[i])
      return i;
  return 0;
}

static int base64_dec_buffer(str,dst,len)
const char *str;void *dst;int len;
{
  int i, j, l;
  unsigned char input[4], output[3], *result = (char *)dst;
  if (str == 0)
    return 0;
  l = strlen(str);
  if (dst == 0 || l > len)
    return (l / 4) * 3;
  memset(dst,0,len);
  for (i=j=0; i<l; i +=4) {
    input[0] = unbase64(str[i]);
    input[1] = unbase64(str[i+1]);
    input[2] = unbase64(str[i+2]);
    input[3] = unbase64(str[i+3]);
    output[0] = (input[0] << 2) | (input[1] >> 4);
    output[1] = (input[1] << 4) | (input[2] >> 2);
    output[2] = (input[2] << 6) | (input[3]);
    result[j] = output[0];
    if (str[i+1] == '=') return j+1;
    result[j+1]=output[1];
    if (str[i+2] == '=') return j+2;
    result[j+2]=output[2];
    j += 3;
  }
  return j;
}

void auth_sigalrm_handler(int sig)
{
	syslog(LOG_NOTICE, "auth server timed out");
	return;
}

/* Check login and password
 * return TRUE in case of failure
 * Put your favorite password check in here
 */

int authenticate(char *login, char *pass)
{
	FILE *fd;
	char line[1000];
	
	fd=ip_open("localhost", 8089);
	if(!fd) return 1;

	fprintf(fd, "smtp\r\n");
	fprintf(fd, "%s\r\n", login);
	fprintf(fd, "%s\r\n", pass);
	fflush(fd);

	alarm(100);
	signal(SIGALRM, auth_sigalrm_handler);

	line[0]=0;	
	fgets(line, sizeof(line)-1, fd);

	fclose(fd);

	return atoi(line)!=100;
}

int auth_exp_delay=1;

int smtp_auth(smtp_t *x, char *p)
{
	char word1[1000], word2[1000], line[1000], 
	     login[2000]="", pass[2000]="";
	     
	if(x->authenticated) {
		echoline(stdout, "235 already authenticated", 0);
		return 0;
		}

        /* netscape 4.5 sends AUTH LOGIN <base64encodedusername>
           microsoft outlook express sends AUTH LOGIN */
           
        get_word(&p, word1, sizeof(word1));

	if(strcasecmp(word1, "LOGIN")==0) {
	        get_word(&p, word2, sizeof(word2));
		if(word2[0]!=0) {
			base64_dec_buffer(word2, login, sizeof(login));
			}
		else {
 			echoline(stdout, "334 VXNlcm5hbWU6", 0); /* b64 <- 'Username:' */
			fflush(stdout);
			if (get_clientinput(line, sizeof(line), x->config->timeout) == NULL)
			  return 1;

			base64_dec_buffer(line, login, sizeof(login));
		}

		echoline(stdout, "334 UGFzc3dvcmQ6", 0); /* b64 <- 'Password:' */

		fflush(stdout);
		if (get_clientinput(line, sizeof(line), x->config->timeout) == NULL)
		  return 1;
		  
		base64_dec_buffer(line, pass, sizeof(pass));

		if(authenticate(login, pass)) {
			echoline(stdout, "535 authentication failed", 0);
			syslog(LOG_NOTICE, "smtp_auth: failed authenticating \"%s\"", login);
			sleep(auth_exp_delay);
			auth_exp_delay*=2;
			return 2;
		}
		else {
			echoline(stdout, "235 Authentication successful", 0);
			syslog(LOG_NOTICE, "smtp_auth: authenticated \"%s\"", login);
			x->authenticated=1;
			return 0;
			}
	}
	else {
		echoline(stdout, "504 Unrecognized authentication type", 0);
		return 0;
	}

	return (0);
}
