#include <iostream>

#include <dblib/os_dep_file.h>
#include <dblib/rfc1123date.h>
#include <dblib/mime_type.h>
#include <dblib/stream_copy.h>
#include <dbclient/auth.h>
#include <dbclient/user.h>
#include <dbclient/logstr.h>
#include <dbclient/client.h>
#include <dbclient/httpd.h>

#include "userhttpd.h"

class userhttpdt:public httpdt
 {
 public:
  userhttpdt(std::istream &_in, std::ostream &_out):
    httpdt(_in, _out) { }

 protected:
  virtual void process_request();

  bool parse_request_url(std::string &section,
                         std::string &user,
                         std::string &file);

  void serve(const std::string &dir,
             const std::string &file);
 };

/*******************************************************************\

Function: userhttpdt::serve

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void userhttpdt::serve(const std::string &dir,
                       const std::string &file)
 {
  std::string filename=build_path(dir, file);

  struct stat statbuf;

  stat(filename.c_str(), &statbuf);

  std::ifstream infile(filename.c_str(), std::ios::in|std::ios::binary);
  if(!infile)
   {
    html << "<html>\n"
            "<head><title>File not found</title></head>\n"
            "<body>File not found: " << url << "</body>\n"
            "</html>\n";
    
    logstr << "File not found: " << url;
    return;
   }

  const std::string mime_type=get_mime_type(filename);

  status_response(200, "OK");

  out << "MIME-Version: 1.0\r\n";
  send_server();
  send_connection();

  if(mime_type!="")
    out << "Content-Type: " << mime_type << "\r\n";

  out << "Content-Length: " << statbuf.st_size << "\r\n"
         "Content-Encoding: binary\r\n"
         "Last-Modified: " << rfc1123date_gmt(statbuf.st_mtime) << "\r\n"
         "\r\n";

  if(filecopy(infile, out, statbuf.st_size))
   {
    /* Error */
   }
 }

/*******************************************************************\

Function: userhttpdt::parse_request_url

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool userhttpdt::parse_request_url(std::string &section,
                                   std::string &user,
                                   std::string &file)
 {
  std::string::size_type pos;

  if(url=="" || url[0]!='/') return TRUE;

  pos=url.find('/', 1);
  if(pos==std::string::npos) return TRUE;
  section=std::string(url, 1, pos);
  file=std::string(url, pos+1);

  pos=file.find('/');
  user=std::string(file, 0, pos);

  return FALSE;
 }

/*******************************************************************\

Function: userhttpdt::process_request

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void userhttpdt::process_request()
 {
  std::string section, user_id, file;

  if(!parse_request_url(section, user_id, file))
   {
    const std::string dir(client.config["section."+section+".dir"]);

    if(dir!="")
     {
      if(client.config.get_bool("section"+section+".auth"))
       {
        if(user_id=="")
          throw auth_excpt("user id required");

        auth("userhttpd");

        if(username!=user_id)
         {
          // maybe an alias
          usert user(user_id);
          if(user["alias"]!=username)
            throw auth_excpt("not authorized");
         }
       }

      if(file.find("..")!=std::string::npos)
       {
        html << "<html>\n"
                "<head><title>Security violation</title></head>\n"
                "<body>Security violation: .. not allowed</body>\n"
                "</html>\n";

        error(404, "Security violation");
       }
      else
       {
        // passed
        serve(dir, file);
       }
     }
   }

  return httpdt::process_request();
 }

/*******************************************************************\

Function: userhttpd

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void userhttpd()
 {
  userhttpdt userhttpd(std::cin, std::cout);

  userhttpd.httpd();
 }
