/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <sys/stat.h>
#include <fstream>

#include "../webonline/www_prof.h"
#include "www_quota.h"

void www_quota(wwwpt &wwwp, const char *filename, bool add)
 {
  struct stat buf;

  if(stat(filename, &buf)==0)
   {
    bool directory=S_ISDIR(buf.st_mode);

    if(add)
     {
      wwwp.set("files", atol(wwwp|"files")+1);

      if(!directory)
        wwwp.set("size", atol(wwwp|"size")+buf.st_blocks*512);
     }
    else
     {
      wwwp.set("files", atol(wwwp|"files")-1);

      if(!directory)
        wwwp.set("size", atol(wwwp|"size")-buf.st_blocks*512);
     }
   }
 }

void add_www_quota_usage(const std::string &name, long &files, long &size)
 {
  wwwpt wwwp;
  wwwp.setname(name);
  wwwp.read();
  size+=atol(wwwp|"size");
  files+=atol(wwwp|"files");
 }

void get_www_quota_usage(userpt &up, long &files, long &size)
 {
  std::string alias_filename;

  files=0;
  size=0;

  add_www_quota_usage(up.name, files, size);

  alias_filename=up.getpath();
  alias_filename+="alias";

  std::ifstream in(alias_filename.c_str());

  if(in==NULL) return;

  char line[TEXTLEN], *tptr, ch;

  while(in.get(line, TEXTLEN-1))
   {
    in.get(ch); // read \n
    tptr=strchr(line, '\n'); if(tptr!=NULL) *tptr=0;
    tptr=strchr(line, '\r'); if(tptr!=NULL) *tptr=0;
    add_www_quota_usage(line, files, size);
   }
 }

void get_max_quota(userpt &up, grouppt &gp, long &files, long &size)
 {
  char tempstr[TEXTLEN];

  getuser(&up, &gp, "homepage.quota.files", tempstr);
  if(tempstr[0]==0)
    files=config.homepage_quota_files;
  else
    files=atol(tempstr);

  getuser(&up, &gp, "homepage.quota.size", tempstr);
  if(tempstr[0]==0)
    size=config.homepage_quota_size;
  else
    size=atol(tempstr);
 }

